import numpy as np
import warnings
from pydantic import BaseModel, field_validator, model_validator, ConfigDict
from typing import Union, Optional, cast

T_Bound = Union[int, float, list, tuple, np.ndarray]


class STPConfig(BaseModel):
    """
    Basically, this model makes the following validation and convertion:

    - ``dim``: a positive integer
    - ``obj``: a positive integer
    - ``lb``: finally a ndarray with shape (dim,)
    - ``ub``: finally a ndarray with shape (dim,)
    - ``fe_init``: a positive integer, default to ``5*dim``
    - ``fe_max``: a positive integer, default to ``11*dim``
    - ``npd``: a positive integer, default to ``1``

    Notes
    -----
    Additionally, this model validate that:
        1. bounds satisfied ``lb<ub`` on each dimension.
        2. ``fe_init<=fe_max``
    """
    model_config = {"arbitrary_types_allowed": True, "extra": "forbid"}
    dim: int
    obj: int
    lb: T_Bound
    ub: T_Bound
    fe_init: int = -1
    fe_max: int = -1
    npd: int = 1

    @field_validator('dim', 'obj')
    def validate_positive_integer(cls, v):
        if v <= 0:
            raise ValueError('dim and obj must be positive integers')
        return v

    @field_validator('fe_init', 'fe_max', 'npd')
    def validate_positive_or_none(cls, v):
        if v < 1:
            raise ValueError(f"Invalid value: {v} (fe_init, fe_max, and npd must be positive)")
        return v

    @model_validator(mode='after')
    def validate_model(self):
        # Convert lb and ub to dim-dimensional arrays if they are int or float
        if isinstance(self.lb, (int, float)):
            self.lb = np.ones(self.dim) * self.lb
        else:
            self.lb = np.asarray(self.lb)

        if isinstance(self.ub, (int, float)):
            self.ub = np.ones(self.dim) * self.ub
        else:
            self.ub = np.asarray(self.ub)

        self.lb = cast(np.ndarray, self.lb)
        self.ub = cast(np.ndarray, self.ub)

        # Check dimensionality
        if self.lb.shape != (self.dim,):
            raise ValueError(f"lb must be a scalar or array of shape ({self.dim},)")
        if self.ub.shape != (self.dim,):
            raise ValueError(f"ub must be a scalar or array of shape ({self.dim},)")

        # Check bounds
        if not np.all(self.lb < self.ub):
            raise ValueError("All elements of lb must be less than corresponding elements of ub")

        # Set default values for fe_init and fe_max
        if self.fe_init == -1:
            self.fe_init = 5 * self.dim
        if self.fe_max == -1:
            self.fe_max = 11 * self.dim

        if self.fe_init > self.fe_max:
            raise ValueError("fe_init must be less than or equal to fe_max")

        return self


class TransformerConfig(BaseModel):
    """
    Basically, this model makes the following validation and conversion:

    - ``dim``: a positive integer (has been validated by STPConfig)
    - ``shift``: finally a ndarray with shape (dim, )
    - ``rotation``: finally a ndarray with shape (dim, dim)
    - ``rotation_inv``: set to inverse of rotation matrix
    """
    model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')
    dim: int
    shift: Optional[np.ndarray] = None
    rotation: Optional[np.ndarray] = None
    rotation_inv: Optional[np.ndarray] = None

    @model_validator(mode='after')
    def check_rotation_and_shift(self):
        if self.rotation is None:
            self.rotation = np.eye(self.dim)
            self.rotation_inv = np.eye(self.dim)
        if self.shift is None:
            self.shift = np.zeros(self.dim)

        if self.rotation.shape != (self.dim, self.dim):
            raise ValueError(f"{self.dim} dimensional task's rotation shape must be ({self.dim}, {self.dim})")
        if self.shift.shape != (self.dim,):
            raise ValueError(f"{self.dim} dimensional task's shift shape must be ({self.dim},)")
        self.rotation = self.rotation.T
        self.rotation_inv = np.linalg.inv(self.rotation.T)

        return self


class FunctionInputs(BaseModel):
    """
    Basically, this model makes the following validation and reshaping:

    - ``x``: finally a ndarray with shape (n, dim)
    - ``dim``: a positive integer (has been validated by STPConfig)

    Notes
    -----
    Additionally, this model validates that:
        1. ``x`` must be a numpy array.
        2. The final shape of ``x`` must be (n, dim).
    """
    model_config = ConfigDict(arbitrary_types_allowed=True, extra='forbid')
    x: np.ndarray
    dim: int

    @model_validator(mode='after')
    def check_and_reshape_x(self):
        if self.x.ndim <= 1:
            self.x = self.x.reshape(-1, self.dim)
        if self.x.ndim != 2:
            raise ValueError(f'x must have shape (n, {self.dim}), got {self.x.shape} instead')
        if self.x.shape[1] != self.dim:
            raise ValueError(f'x must have shape (n, {self.dim}), got {self.x.shape} instead')
        return self


class PlottingArgs(BaseModel):
    """
    This model will make sure:

    - ``dim:int>=2``
    - ``0<=dims[0]<dims[1]<dim``
    - ``10<=n_points:int<=1000``
    - ``lb<=fixed:ndarray<=ub``, ``fixed.shape=(dim,)``
    """
    model_config = ConfigDict(arbitrary_types_allowed=True)
    dim: int
    dims: tuple
    n_points: int
    lb: np.ndarray
    ub: np.ndarray
    fixed: Union[int, float, np.ndarray, None]

    @field_validator('dim')
    def dim_must_be_ge_2(cls, v):
        if v < 2:
            raise ValueError(f"only support dim>=2 functions, got dim={v}")
        return v

    @field_validator('dims')
    def dims_must_ordered_and_different(cls, v):
        if len(v) != 2:
            raise ValueError(f"dims must be a tuple of two >=0 integers, got {v}")
        if min(v) < 0:
            raise ValueError("dims must be integers and >= 0")
        if v[0] == v[1]:
            raise ValueError(f"dims should be two different integers, got dims={v}")
        return min(v), max(v)

    @field_validator('n_points')
    def warning_if_too_large(cls, v):
        if v > 1000:
            warnings.warn("A large n_points may cause slow plotting. Using 1000 instead.")
            return 1000
        if v < 10:
            warnings.warn("A small n_points may cause detail loss in plotting. Using 10 instead.")
            return 10
        return v

    @model_validator(mode='after')
    def check_and_set_defaults(self):
        if self.dims[1] >= self.dim:
            raise ValueError(f"selected dim must be in [0, {self.dim - 1}], got dims={self.dims}")
        if self.fixed is None:
            self.fixed = (self.lb + self.ub) / 2
        elif isinstance(self.fixed, (int, float)):
            self.fixed = np.ones(self.dim) * self.fixed
        self.fixed = cast(np.ndarray, self.fixed)
        if self.fixed.shape != (self.dim,):
            raise ValueError(f"fixed shape, if a ndarray, must be ({self.dim},), got fixed shape={self.fixed.shape}")
        if np.any(self.fixed < self.lb) or np.any(self.fixed > self.ub):
            raise ValueError(f"fixed point must be within [lb, ub], got \n"
                             f"fixed={self.fixed}\nlb={self.lb}\nub={self.ub}")
        return self



__version__ = "0.2.2"

from .experiments import Launcher, Reports
from .utilities.loaders import (
    list_algorithms,
    load_algorithm,
    list_problems,
    init_problem,
    load_problem,
)

__all__ = [
    'Reports',
    'Launcher',
    'init_problem',
    'load_problem',
    'list_problems',
    'load_algorithm',
    'list_algorithms',
]



import numpy as np
from typing import Union
from numpy import ndarray
from tabulate import tabulate

from pyfmto.utilities.schemas import STPConfig

__all__ = ['Solution']


def _only_single_obj(func):
    def wrapped(*args, **kwargs):
        instance = args[0]
        if instance.obj != 1:
            raise AttributeError(f"Expected problem obj=1, got problem obj={instance.obj}")
        return func(*args, **kwargs)

    return wrapped


class Solution:
    _dim: int
    _obj: int
    _fe_init: int
    _fe_max: int
    _npd: int
    _prev_size: int
    _lb: ndarray
    _ub: ndarray
    _x: ndarray
    _y: ndarray
    _x_global: ndarray
    _y_global: ndarray

    def __init__(self, data: Union[STPConfig, dict]):
        self._prev_size = 0
        if isinstance(data, dict):
            self._init_from_dict(data)
        elif isinstance(data, STPConfig):
            self._init_from_config(data)
        else:
            raise TypeError(f"Solution() must be initialized with a dict or STPConfig, got {type(data)} instead.")

    def __repr__(self):
        return f"SolutionSet(dim={self.dim}, obj={self.obj}, solution_size={self.size})"

    def __str__(self):
        solution_info = {
            'dim': [self.dim],
            'obj': [self.obj],
            'init': [self.fe_init],
            'total': [self.size]
        }
        return tabulate(solution_info, headers='keys', tablefmt='rounded_grid')

    def to_dict(self):
        return self.__dict__

    def _init_from_dict(self, data: dict):
        required_keys = set(self.required_keys)
        actual_keys = set(data.keys())
        missing_keys = required_keys - actual_keys
        extra_keys = actual_keys - required_keys
        if missing_keys or extra_keys:
            error_msg = "Invalid data:\n"
            error_msg += f"  Expected keys: {sorted(required_keys)}"
            error_msg += f"  Missing keys: {sorted(missing_keys)}\n"
            error_msg += f"  Extra keys: {sorted(extra_keys)}\n"
            raise ValueError(error_msg)
        self.__dict__.update(data.copy())

    def _init_from_config(self, config: STPConfig):
        self._dim = config.dim
        self._obj = config.obj
        self._fe_init = config.fe_init
        self._fe_max = config.fe_max
        self._npd = config.npd
        self._lb = config.lb  # type: ignore
        self._ub = config.ub  # type: ignore
        self._x = np.array([])
        self._y = np.array([])
        self._x_global = np.array([])
        self._y_global = np.array([])

    @property
    def required_keys(self):
        return sorted(self.__class__.__annotations__.keys())

    @property
    def size(self):
        return self._x.shape[0]

    @property
    def num_updated(self):
        num = self.size - self._prev_size
        self._prev_size = self.size
        return num

    @property
    def x(self) -> ndarray:
        return self._x

    @property
    def y(self) -> ndarray:
        return self._y

    @property
    def x_init(self) -> ndarray:
        return self._x[:self.fe_init]

    @property
    def y_init(self) -> ndarray:
        return self._y[:self.fe_init]

    @property
    def x_alg(self) -> ndarray:
        return self._x[self.fe_init:]

    @property
    def y_alg(self) -> ndarray:
        return self._y[self.fe_init:]

    @property
    def x_global(self) -> ndarray:
        return self._x_global

    @property
    def y_global(self) -> ndarray:
        return self._y_global

    @property
    def initialized(self):
        return self.size >= self.fe_init

    @property
    def dim(self):
        return self._dim

    @property
    def obj(self):
        return self._obj

    @property
    def lb(self):
        return self._lb

    @property
    def ub(self):
        return self._ub

    @property
    def fe_init(self):
        return self._fe_init

    @property
    def fe_max(self):
        return self._fe_max

    @property
    def npd(self) -> int:
        return self._npd

    @property
    @_only_single_obj
    def y_max(self):
        return np.max(self.y)

    @property
    @_only_single_obj
    def y_min(self):
        return np.min(self.y)

    @property
    @_only_single_obj
    def x_of_y_max(self):
        idx = np.argmax(self.y.flatten())
        return self.x[idx]

    @property
    @_only_single_obj
    def x_of_y_min(self):
        idx = np.argmin(self.y.flatten())
        return self.x[idx]

    @property
    @_only_single_obj
    def y_homo_decrease(self):
        y_arr = self.y.flatten()
        res = [y_arr[0]]
        for y in y_arr[1:]:
            if y <= res[-1]:
                res.append(y)
            else:
                res.append(res[-1])
        return np.array(res)

    @property
    @_only_single_obj
    def y_homo_increase(self):
        y_arr = self.y.flatten()
        res = [y_arr[0]]
        for y in y_arr[1:]:
            if y >= res[-1]:
                res.append(y)
            else:
                res.append(res[-1])
        return np.array(res)

    def append(self, x: ndarray, y: ndarray):
        """
        Append one or more solutions to the solution set.
        """
        x = np.array(x)
        y = np.array(y)
        if x.ndim != 2:
            x = x.reshape(-1, self.dim)
        if y.ndim != 2:
            y = y.reshape(-1, self.obj)
        if x.shape[0] != y.shape[0]:
            raise ValueError("expect x,y to have same number of rows, "
                             f"got (x {x.shape[0]}, y {y.shape[0]}) instead.")

        if self.size == 0:
            self._x = x
            self._y = y
        else:
            self._x = np.vstack((self.x, x))
            self._y = np.vstack((self.y, y))



"""
Single-Task Single-Objective Benchmark Problems

This module implements a collection of classical single-objective continuous optimization benchmark functions,
commonly used in evolutionary multitasking optimization, federated optimization, and other multitask learning scenarios.

Each function is defined as a class derived from the base class `SingleTaskProblem`, and overrides the `evaluate` method
to support batched input and vectorized computation using NumPy.

The following benchmark problems are included:
    - Griewank
    - Rastrigin
    - Ackley
    - Schwefel
    - Sphere
    - Rosenbrock
    - Weierstrass
    - Ellipsoid

References
----------
Definition:
    B. Da, Y.-S. Ong, L. Feng, A. K. Qin, A. Gupta, Z. Zhu, C.-K. Ting,
    K. Tang, and X. Yao, "Evolutionary multitasking for single-objective
    continuous optimization: Benchmark problems, performance metric, and
    baseline results," arXiv preprint arXiv:1706.03470, 2017.

Implementation:
    Zhu, H., Wang, X., & Jin, Y. (2023). Federated Many-Task Bayesian
    Optimization. IEEE Transactions on Evolutionary Computation, 1–1.
    https://doi.org/10.1109/TEVC.2023.3279775
"""

import numpy as np

from ..problem import SingleTaskProblem as Stp

__all__ = [
    "Ackley",
    "Sphere",
    "Schwefel",
    "Griewank",
    "Ellipsoid",
    "Rastrigin",
    "Rosenbrock",
    "Weierstrass",
]


class Griewank(Stp):

    def __init__(self, dim=10, lb=-600, ub=600, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)

    def _eval_single(self, x):
        i = np.arange(self.dim) + 1
        f1 = np.sum(x ** 2 / 4000)
        f2 = np.prod(np.cos(x / np.sqrt(i)))
        out = f1 - f2 + 1
        return out


class Rastrigin(Stp):

    def __init__(self, dim=10, lb=-5, ub=5, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)

    def _eval_single(self, x):
        square_term = x ** 2
        cos_term = 10 * np.cos(2 * np.pi * x)
        out = np.sum(square_term - cos_term + 10)
        return out


class Ackley(Stp):

    def __init__(self, dim=10, lb=-32.768, ub=32.768, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)

    def _eval_single(self, x):
        sum1 = np.sum(x ** 2)
        sum2 = np.sum(np.cos(2 * np.pi * x))
        out = -20 * np.exp(-0.2 * np.sqrt(sum1 / self.dim)) - np.exp(sum2 / self.dim) + 20 + np.e
        return out


class Schwefel(Stp):

    def __init__(self, dim=10, lb=-500, ub=500, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)
        self.set_x_global(np.ones(self.dim) * 420.9687)

    def _eval_single(self, x):
        term = x * np.sin(np.sqrt(np.abs(x)))
        out = 418.9829 * self.dim - np.sum(term)
        return out


class Sphere(Stp):

    def __init__(self, dim=10, lb=-100, ub=100, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)

    def _eval_single(self, x):
        return np.sum(x ** 2)


class Rosenbrock(Stp):
    def __init__(self, dim=10, lb=-2.048, ub=2.048, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)
        self.set_x_global(np.ones(self.dim))

    def _eval_single(self, x):
        x_squared_diff = x[1:] ** 2 - x[:-1]
        term1 = 100 * np.sum(np.power(x_squared_diff, 2))
        term2 = np.sum(np.power(x[:-1] - 1, 2))
        return term1 + term2


class Weierstrass(Stp):

    def __init__(self, dim=10, lb=-0.5, ub=0.5, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)

    def _eval_single(self, x):
        a = 0.5
        b = 3
        kmax = 21

        k_values = np.arange(kmax)
        a_k = a ** k_values
        b_k = b ** k_values

        term1 = np.sum(a_k * np.cos(2 * np.pi * b_k * (x[:, np.newaxis] + 0.5)), axis=1)
        term2 = np.sum(a_k * np.cos(2 * np.pi * b_k * 0.5), axis=0)

        obj = np.sum(term1) - self.dim * term2

        return obj


class Ellipsoid(Stp):

    def __init__(self, dim=10, lb=-5.12, ub=5.12, **kwargs):
        super().__init__(dim=dim, obj=1, lb=lb, ub=ub, **kwargs)

    def _eval_single(self, x):
        d_l = np.arange(self.dim) + 1
        out = np.sum(d_l * x ** 2)
        return out




from .arxiv2017 import Arxiv2017
from .tevc2024 import Tevc2024
from .tetci2019 import Tetci2019
from .cec2022 import Cec2022

__all__ = [
    'Arxiv2017',
    'Tevc2024',
    'Tetci2019',
    'Cec2022',
]



from collections import defaultdict
from enum import Enum, auto
from pydantic import validate_call
from typing import Optional, Any

from pyfmto.utilities import logger

__all__ = ['Actions', 'ClientPackage', 'SyncDataManager']


class Actions(Enum):
    REGISTER = auto()
    QUIT = auto()


class ClientPackage:
    def __init__(self, cid: Optional[int], action: Any):
        self.cid = cid
        self.action = action


class SyncDataManager:
    def __init__(self):
        self._source: dict[int, dict[int, Any]] = defaultdict(dict)
        self._result: dict[int, dict[int, Any]] = defaultdict(dict)

    @validate_call
    def update_src(self, cid: int, version: int, data: Any) -> None:
        """
        Update aggregation source of a client
        """
        self._source[cid][version] = data

    @validate_call
    def update_res(self, cid: int, version: int, data: Any) -> None:
        """
        Update aggregation results of a client
        """
        self._result[cid][version] = data

    @validate_call
    def lts_src_ver(self, cid: int) -> int:
        """
        Return the latest source version of a client
        """
        data = self._source.get(cid, {-1: None})
        return max(data.keys())

    @validate_call
    def lts_res_ver(self, cid: int) -> int:
        """
        Return the latest aggregation result version of a client
        """
        data = self._result.get(cid, {-1: None})
        return max(data.keys())

    @validate_call
    def get_src(self, cid: int, version: int) -> Any:
        """
        - Return the aggregation source
        - Return None if not found
        """
        if cid not in self._source:
            logger.debug(f"cid={cid} not found in source")
            return None
        data = self._source[cid].get(version)
        if data is None:
            logger.debug(f"cid={cid} version={version} not found in source")
        return data

    @validate_call
    def get_res(self, cid: int, version: int) -> Any:
        """
        - Return the aggregation result data
        - Return None if not found
        """
        if cid not in self._result:
            logger.debug(f"cid={cid} not found in result")
            return None
        data = self._result[cid].get(version)
        if data is None:
            logger.debug(f"cid={cid} version={version} not found in result")
        return data

    @property
    def available_src_ver(self) -> int:
        """
        - The latest consistent version uploaded by all clients
        - -1 if no source data is available
        """
        try:
            return min([max(data.keys()) for data in self._source.values()])
        except ValueError:
            return -1

    @property
    def num_clients(self) -> int:
        """
        The number of clients that have uploaded least 1 source data
        """
        return len(self._source)



import numpy as np
from pathlib import Path
from scipy.io import loadmat

from ..problem import MultiTaskProblem as Mtp
from ..benchmarks import (
    Ackley,
    Griewank,
    Rastrigin,
    Rosenbrock,
    Schwefel,
    Sphere,
    Weierstrass)

__all__ = ["Arxiv2017"]


class Arxiv2017(Mtp):
    """
    dim: 10  # 1 <= dim <= 50
    """
    is_realworld = False
    intro = """
        A many-task single-objective optimization problem based on transformed versions
        of classical single-objective benchmark functions. Each task corresponds to an
        individual optimization problem.
    """
    notes = """
        The number of tasks is either 17 or 18, depending on the input dimension.
        If dim <= 25, 18 tasks, If dim > 25, 17 tasks (due to Weierstrass function only support 1 <= dim <= 25)
        All tasks share the same dimensionality.
        Tasks differ by rotation matrices and shift vectors applied to their base functions.
        This benchmark was used in federated many-task Bayesian optimization experiments.
        Task10 same with task13. Task6 same with task18.
    """
    references = [
        """
        B. Da, Y.-S. Ong, L. Feng, A. K. Qin, A. Gupta, Z. Zhu, C.-K. Ting,
        K. Tang, and X. Yao, "Evolutionary multitasking for single-objective
        continuous optimization: Benchmark problems, performance metric, and
        baseline results," arXiv preprint arXiv:1706.03470, 2017.
        """,
        """
        Zhu, H., Wang, X., & Jin, Y. (2023). Federated Many-Task Bayesian
        Optimization. IEEE Transactions on Evolutionary Computation, 1–1.
        https://doi.org/10.1109/TEVC.2023.3279775
        """
    ]

    def __init__(self, dim=10, **kwargs):
        if not 1 <= dim <= 50:
            raise ValueError(f"dim must be in [1, 50], but got {dim}")
        super().__init__(dim, **kwargs)

    def get_info(self):
        return {
            "TaskID": [t.id for t in self._problem],
            "TaskName": [t.name for t in self._problem],
            "DecDim": [t.dim for t in self._problem],
            "Lower": [t.lb[0] for t in self._problem],
            "Upper": [t.ub[0] for t in self._problem]
        }

    def _init_tasks(self, dim, **kwargs):
        datasets = Path(__file__).parents[1] / 'datasets' / 'arxiv2017'
        cih = loadmat(str(datasets / 'CI_H.mat'))
        cim = loadmat(str(datasets / 'CI_M.mat'))
        cil = loadmat(str(datasets / 'CI_L.mat'))
        pih = loadmat(str(datasets / 'PI_H.mat'))
        pim = loadmat(str(datasets / 'PI_M.mat'))
        pil = loadmat(str(datasets / 'PI_L.mat'))
        nih = loadmat(str(datasets / 'NI_H.mat'))
        nim = loadmat(str(datasets / 'NI_M.mat'))
        nil = loadmat(str(datasets / 'NI_L.mat'))

        rot1, rot2 = cih['Rotation_Task1'][0:dim, 0:dim], cih['Rotation_Task2'][0:dim, 0:dim]
        rot3, rot4 = cim['Rotation_Task1'][0:dim, 0:dim], cim['Rotation_Task2'][0:dim, 0:dim]
        rot5, rot6 = cil['Rotation_Task1'][0:dim, 0:dim], None
        rot7, rot8 = pih['Rotation_Task1'][0:dim, 0:dim], None
        rot9, rot10 = pim['Rotation_Task1'][0:dim, 0:dim], None
        rot11, rot12 = pil['Rotation_Task1'][0:dim, 0:dim], pil['Rotation_Task2'][0:dim, 0:dim]
        rot13, rot14 = None, nih['Rotation_Task2'][0:dim, 0:dim]
        rot15, rot16 = nim['Rotation_Task1'][0:dim, 0:dim], nim['Rotation_Task2'][0:dim, 0:dim]
        rot17, rot18 = nil['Rotation_Task1'][0:dim, 0:dim], None

        shift1, shift2 = np.squeeze(cih['GO_Task1'])[0:dim], np.squeeze(cih['GO_Task2'])[0:dim]
        shift3, shift4 = np.squeeze(cim['GO_Task1'])[0:dim], np.squeeze(cim['GO_Task2'])[0:dim]
        shift5, shift6 = np.squeeze(cil['GO_Task1'])[0:dim], None
        shift7, shift8 = np.squeeze(pih['GO_Task1'])[0:dim], np.squeeze(pih['GO_Task2'])[0:dim]
        shift9, shift10 = np.squeeze(pim['GO_Task1'])[0:dim], None
        shift11, shift12 = np.squeeze(pil['GO_Task1'])[0:dim], np.squeeze(pil['GO_Task2'])[0:dim]
        shift13, shift14 = None, np.squeeze(nih['GO_Task2'])[0:dim]
        shift15, shift16 = np.squeeze(nim['GO_Task1'])[0:dim], np.squeeze(nim['GO_Task2'])[0:dim]
        shift17, shift18 = np.squeeze(nim['GO_Task1'])[0:dim], None

        func1 = Griewank(dim, lb=-100, ub=100, **kwargs)
        func2 = Rastrigin(dim, lb=-50, ub=50, **kwargs)
        func3 = Ackley(dim, lb=-50, ub=50, **kwargs)
        func4 = Rastrigin(dim, lb=-50, ub=50, **kwargs)
        func5 = Ackley(dim, lb=-50, ub=50, **kwargs)
        func6 = Schwefel(dim, lb=-500, ub=500, **kwargs)
        func7 = Rastrigin(dim, lb=-50, ub=50, **kwargs)
        func8 = Sphere(dim, lb=-100, ub=100, **kwargs)
        func9 = Ackley(dim, lb=-50, ub=50, **kwargs)
        func10 = Rosenbrock(dim, lb=-50, ub=50, **kwargs)
        func11 = Ackley(dim, lb=-50, ub=50, **kwargs)
        func12 = Weierstrass(dim, lb=-0.5, ub=0.5, **kwargs)
        func13 = Rosenbrock(dim, lb=-50, ub=50, **kwargs)
        func14 = Rastrigin(dim, lb=-50, ub=50, **kwargs)
        func15 = Griewank(dim, lb=-100, ub=100, **kwargs)
        func16 = Weierstrass(dim, lb=-0.5, ub=0.5, **kwargs)
        func17 = Rastrigin(dim, lb=-50, ub=50, **kwargs)
        func18 = Schwefel(dim, lb=-500, ub=500, **kwargs)

        functions = [
            func1, func2, func3, func4, func5, func6, func7, func8, func9,
            func10, func11, func12, func13, func14, func15, func16, func17, func18
        ]

        # Set id
        for f_id in range(len(functions)):
            functions[f_id].set_id(f_id + 1)

        rot_mats = [
            rot1, rot2, rot3, rot4, rot5, rot6, rot7, rot8, rot9,
            rot10, rot11, rot12, rot13, rot14, rot15, rot16, rot17, rot18
        ]

        shift_mats = [
            shift1, shift2, shift3, shift4, shift5, shift6, shift7, shift8, shift9,
            shift10, shift11, shift12, shift13, shift14, shift15, shift16, shift17, shift18
        ]

        if dim > 25:
            # Transform matrix of pil 'Rotation_Task2' and 'GO_Task2' with different shape (25,25)
            del functions[11]
            del rot_mats[11]
            del shift_mats[11]

        for f, rot_mat, shift_mat in zip(functions, rot_mats, shift_mats):
            f.set_transform(rot_mat, shift_mat)

        return functions



import copy
import numpy as np
import os
import pandas as pd
import seaborn
import shutil
import time
from collections import defaultdict, namedtuple
from PIL import Image
from matplotlib import pyplot as plt
from scipy import stats
from pathlib import Path
from typing import Optional, Union

from pyfmto.problems import Solution
from pyfmto.utilities import logger, save_msgpack, titled_tabulate, load_msgpack, colored

StatisData = namedtuple("StatisData", ['mean', 'std', 'se', 'opt'])

__all__ = [
    "MetaData",
    "RunSolutions",
    "ClientDataStatis",
    "MergedResults",
    "ReporterUtils"
]


class RunSolutions:
    def __init__(self, run_solutions: Optional[dict] = None):
        self._solutions: dict[int, dict] = {}
        if run_solutions:
            self.__dict__.update(copy.deepcopy(run_solutions))

    def update(self, cid: int, solution: Solution):
        self._solutions[cid] = copy.deepcopy(solution.to_dict())

    @property
    def solutions(self) -> list[Solution]:
        return [Solution(self._solutions[cid]) for cid in self.sorted_ids]

    def __getitem__(self, item):
        try:
            return Solution(self._solutions[item])
        except KeyError:
            raise KeyError(f"Client id '{item}' not found in results")

    def items(self) -> list[tuple[int, Solution]]:
        return list(zip(self.sorted_ids, self.solutions))

    def to_msgpack(self, filename: Union[str, Path] = 'results.msgpack'):
        if self.num_clients == 0:
            raise ValueError('Empty RunSolutions')
        else:
            data = copy.deepcopy(self.__dict__)
            save_msgpack(data, filename)
            logger.info(f"Results saved to {filename}")

    @property
    def num_clients(self):
        return len(self._solutions)

    @property
    def sorted_ids(self):
        return sorted(map(int, self._solutions.keys()))


class ClientDataStatis:
    fe_init: int
    fe_max: int
    lb: np.ndarray
    ub: np.ndarray
    x_init: np.ndarray
    y_init: np.ndarray
    x_alg: np.ndarray
    y_alg: np.ndarray
    y_dec_mat: np.ndarray
    y_inc_mat: np.ndarray
    x_global: np.ndarray
    y_global: np.ndarray
    y_mean: np.ndarray

    def __init__(self, solutions: list[Solution]):
        if len(solutions) < 1:
            raise ValueError('Empty list of Solutions')
        self._preprocessing(solutions)

    def _preprocessing(self, solutions):
        x = []
        y_dec = []
        y_inc = []
        x_init = []
        y_init = []
        x_alg = []
        y_alg = []
        for solution in solutions:
            x.append(solution.x)
            y_dec.append(solution.y_homo_decrease)
            y_inc.append(solution.y_homo_increase)
            x_init.append(solution.x_init)
            y_init.append(solution.y_init)
            x_alg.append(solution.x_alg)
            y_alg.append(solution.y_alg)
        self.x_init = np.vstack(x_init)
        self.y_init = np.vstack(y_init)
        self.x_alg = np.vstack(x_alg)
        self.y_alg = np.vstack(y_alg)
        self.y_dec_mat = np.array(y_dec)
        self.y_inc_mat = np.array(y_inc)
        self.y_dec_mat[self.y_dec_mat < 1e-20] = 1e-20
        self.y_inc_mat[self.y_inc_mat < 1e-20] = 1e-20
        self.x_global = solutions[0].x_global
        self.y_global = solutions[0].y_global
        self.fe_init = solutions[0].fe_init
        self.fe_max = solutions[0].fe_max
        self.lb = solutions[0].lb
        self.ub = solutions[0].ub

    @property
    def is_known_optimal(self):
        return self.x_global.size > 0

    @property
    def y_dec_statis(self) -> StatisData:
        return ReporterUtils.get_mat_statis(self.y_dec_mat)

    @property
    def y_inc_statis(self) -> StatisData:
        return ReporterUtils.get_mat_statis(self.y_inc_mat)

    @property
    def y_dec_log_statis(self) -> StatisData:
        return ReporterUtils.get_mat_statis(np.log10(self.y_dec_mat))

    @property
    def y_inc_log_statis(self) -> StatisData:
        return ReporterUtils.get_mat_statis(np.log10(self.y_inc_mat))


class MergedResults:

    def __init__(self, runs_data: list[RunSolutions]):
        if len(runs_data) < 1:
            raise ValueError('Empty list of RunSolutions')
        self._original: list[RunSolutions] = runs_data
        self._reorganized: dict[int, list[Solution]] = defaultdict(list)
        self._merged: dict[str, ClientDataStatis] = {}
        self._reorganizing()
        self._merging()

    def _reorganizing(self):
        for run_data in self._original:
            for cid, solution in run_data.items():
                self._reorganized[cid].append(solution)

    def _merging(self):
        self._merged = {
            f"Client {cid:>02d}": ClientDataStatis(solutions)
            for cid, solutions in self._reorganized.items()
        }

    def get_statis(self, c_name) -> ClientDataStatis:
        return self._merged[c_name]

    def items(self) -> list[tuple[str, ClientDataStatis]]:
        return list(self._merged.items())

    @property
    def num_runs(self):
        return len(self._original)

    @property
    def dim(self):
        return self._original[0].solutions[0].dim

    @property
    def sorted_names(self):
        return sorted(self._merged.keys())


class MetaData:

    def __init__(
            self,
            data: dict[str, MergedResults],
            problem: str,
            npd: str,
            root: Path):
        self.data = data
        self.problem = problem
        self.npd = npd
        self.root = root

    def __len__(self):
        return len(self.data)

    def __getitem__(self, item):
        return self.data[item]

    def items(self) -> list[tuple[str, MergedResults]]:
        return list(self.data.items())

    def is_empty(self):
        return len(self.data) == 0

    @property
    def clt_names(self):
        if self.alg_num == 0:
            return []
        else:
            return list(self.data.values())[0].sorted_names

    @property
    def clt_num(self):
        return len(self.clt_names)

    @property
    def alg_num(self):
        return len(self.data)

    @property
    def dim(self):
        if self.alg_num == 0:
            return 0
        else:
            return list(self.data.values())[0].dim

    @property
    def num_runs(self):
        return list(self.data.values())[0].num_runs

    @property
    def alg_names(self):
        return list(self.data.keys())

    @property
    def report_filename(self) -> Path:
        if self.is_empty():
            raise ValueError(f"Empty data for analysis on [{self.problem}] with [{self.npd}]")
        else:
            filedir = self.root / f"{time.strftime('%Y-%m-%d')}" / f"{self.alg_names[-1]}" / f"{self.problem}"
            filedir.mkdir(parents=True, exist_ok=True)
            file_name = filedir / self.npd
            return file_name


class ReporterUtils:

    @staticmethod
    def check_rows(data_list: list[list], col_title: list[str], row_title: list[str], msg=''):
        try:
            data_mat = np.array(data_list, dtype=int)
        except Exception:
            tab = titled_tabulate(msg, '=', data_list, headers=col_title, showindex=row_title, tablefmt='psql')
            raise ValueError(f"\n{tab}")
        if not np.all(np.equal(data_mat, data_mat[0])):
            tab = titled_tabulate(msg, '=', data_list, headers=col_title, showindex=row_title, tablefmt='psql')
            raise ValueError(f"\n{tab}")

    @staticmethod
    def find_grid_shape(size):
        """
        Determine the shape of a grid (a, b) for a given size such that:
        1. a * b >= size
        2. a * b - size is minimized
        3. a / b is as close to 1 as possible

        Parameters
        ----------
        size : int
            The total number of elements in the grid.

        Returns
        -------
        tuple
            A tuple (a, b) representing the dimensions of the grid.
            a and b are integers such that a * b >= size and a / b is minimized.
        """
        if size <= 0:
            raise ValueError("Size must be a positive integer.")
        a = int(np.sqrt(size))
        b = size // a
        while a * b < size:
            if a < b:
                a += 1
            else:
                b += 1

        return (a, b) if a > b else (b, a)

    @staticmethod
    def get_t_test_suffix(opt_list1, opt_list2, mean1, mean2, pvalue):
        diff = mean1 - mean2
        _, p = stats.ttest_ind(opt_list1, opt_list2)
        if p > pvalue:
            suffix = "≈"
        elif diff > 0:
            suffix = "-"
        else:
            suffix = "+"
        return suffix

    @staticmethod
    def plotting(ax, merged_results: MergedResults, alg_name, c_name, showing_size, in_log_scale, alpha, color):
        c_data = merged_results.get_statis(c_name)
        fe_init = c_data.fe_init
        fe_max = c_data.fe_max
        size_optimization = fe_max - fe_init
        size_plot = fe_max if showing_size == -1 else showing_size
        x_indices: np.ndarray = np.arange(size_plot) + 1
        avg = c_data.y_dec_log_statis.mean[-size_plot:] if in_log_scale else c_data.y_dec_statis.mean[-size_plot:]
        se = c_data.y_dec_log_statis.se[-size_plot:] if in_log_scale else c_data.y_dec_statis.se[-size_plot:]
        se_upper = list(map(lambda x: x[0] - x[1], zip(avg, se)))
        se_lower = list(map(lambda x: x[0] + x[1], zip(avg, se)))
        ax.plot(x_indices, avg, label=alg_name, color=color)
        ax.fill_between(x_indices, se_upper, se_lower, alpha=alpha, color=color)
        if size_plot > size_optimization:
            return size_plot - size_optimization
        else:
            return 0

    @staticmethod
    def load_runs_data(file_dir: Path, prefix: str = '') -> list[RunSolutions]:
        """
        Load all RunSolutions from a directory.
        Parameters
        ----------
        file_dir: Path
            The directory to load RunSolutions from.
        prefix: str
            The prefix of the RunSolutions file.
        """
        suffix = '.msgpack'
        if file_dir.exists():
            filenames = [
                file_dir / f_name
                for f_name in os.listdir(file_dir) if f_name.startswith(prefix) and f_name.endswith(suffix)
            ]
            return [RunSolutions(load_msgpack(p)) for p in filenames]
        else:
            return []

    @staticmethod
    def get_optimality_mask_mat(mean_dict):
        df = pd.DataFrame.from_dict(mean_dict)
        data_mat = df.to_numpy()
        mat_shape = data_mat.shape[0], data_mat.shape[1]
        mask = np.zeros(shape=mat_shape, dtype=bool)

        row, col = mask.shape
        min_idx = np.argmin(data_mat, axis=1)
        for i in range(row):
            mask[i, min_idx[i]] = True

        # row+1: the last row is the count of best solution of each algorithm
        # col+1: the first col is the index of Client
        add_row = np.zeros(shape=mask.shape[1], dtype=bool)
        mask = np.vstack((mask, add_row))
        add_col = np.zeros(shape=mask.shape[0], dtype=bool).reshape(-1, 1)
        mask = np.hstack((add_col, mask))
        return mask

    @staticmethod
    def get_mat_statis(y_mat: np.ndarray):
        rows = y_mat.shape[0]
        mean = np.mean(y_mat, axis=0)
        std = np.std(y_mat, ddof=1, axis=0)
        se = std / np.sqrt(rows)
        opt = y_mat[:, -1]
        return StatisData(mean, std, se, opt)

    @staticmethod
    def plot_violin(statis: ClientDataStatis, figsize, filename: Path, title: str, dpi: float):
        n_dims = statis.x_init.shape[1]
        df_init = pd.DataFrame(statis.x_init, columns=[f'x{i + 1}' for i in range(n_dims)])
        df_optimized = pd.DataFrame(statis.x_alg, columns=[f'x{i + 1}' for i in range(n_dims)])
        hue = 'X From'
        df_init[hue] = 'Init'
        df_optimized[hue] = 'Alg'
        df_combined = pd.concat([df_init, df_optimized], ignore_index=True)
        df_melted = df_combined.melt(id_vars=[hue], var_name='Dimension', value_name='Value')
        plt.figure(figsize=figsize, dpi=dpi)
        ax = seaborn.violinplot(
            data=df_melted,
            x='Dimension',
            y='Value',
            hue=hue,
            palette=['#ff7f50', '#2d98da'],
            split=True,
            inner='quartile',
            linewidth=0.5
        )
        x_global = statis.x_global
        if x_global.size > 0:
            for dim in range(n_dims):
                ax.plot(dim, x_global[dim], 'r*', markersize=8, markeredgecolor='w', linewidth=0.5)
        plt.title(title)
        plt.xlabel('Dimension Index')
        plt.ylabel('Value')
        plt.tight_layout()
        plt.savefig(filename)
        plt.close()

    @staticmethod
    def merge_images_in(file_dir: Path, clear: bool):
        file_paths = list(file_dir.iterdir())
        suffix = file_paths[0].suffix
        n_row, n_col = ReporterUtils.find_grid_shape(len(file_paths))
        merge_from = file_dir.name
        sorted_paths = sorted(str(path) for path in file_paths)
        images = [Image.open(path) for path in sorted_paths]
        sizes = np.array([img.size for img in images])
        width, height = np.min(sizes, axis=0)
        canvas_width = width * n_col
        canvas_height = height * n_row
        merged = Image.new('RGB', (canvas_width, canvas_height), color='white')

        for i, img in enumerate(images):
            row_id = i // n_col
            col_id = i % n_col
            resized_img = img.resize((width, height), Image.Resampling.LANCZOS)
            merged.paste(resized_img, (col_id * width, row_id * height))

        merged.save(file_dir.parent / f'{merge_from}{suffix}')
        if clear:
            shutil.rmtree(file_dir)

    @staticmethod
    def check_suffix(suffix: str, merge: bool):
        if merge and (suffix not in ['.png', '.jpg']):
            print(f"Only support suffix {colored('.png or .jpg', 'green')} "
                  f"when {colored('merge is True', 'green')}, defaulted to '.png'")
            return '.png'
        else:
            return suffix



from .loggers import logger

from .tools import (
    colored,
    clear_console,
    titled_tabulate,
    tabulate_formats,
    matched_str_head,
    terminate_popen,
)

from .stroptions import (
    Cmaps,
    StrColors,
    SeabornStyles,
    SeabornPalettes,
)

from .io import (
    load_yaml,
    save_yaml,
    parse_yaml,
    dumps_yaml,
    load_msgpack,
    save_msgpack,
)

__all__ = [
    'logger',
    'colored',
    'clear_console',
    'terminate_popen',
    'titled_tabulate',
    'matched_str_head',
    'tabulate_formats',
    'Cmaps',
    'StrColors',
    'SeabornStyles',
    'SeabornPalettes',
    'load_yaml',
    'save_yaml',
    'parse_yaml',
    'dumps_yaml',
    'load_msgpack',
    'save_msgpack',
]



from pathlib import Path
from scipy.io import loadmat
from typing import Literal, Type

from .. import benchmarks
from ..problem import (MultiTaskProblem as Mtp, SingleTaskProblem)

__all__ = ["Tevc2024"]

T_SrcProblem = Literal[
    'Griewank', 'Rastrigin', 'Ackley', 'Schwefel', 'Sphere', 'Rosenbrock', 'Weierstrass', 'Ellipsoid']


class Tevc2024(Mtp):
    """
    dim: 10  # 1 <= dim <= 10
    src_problem: Ackley  # Options: Ackley, Griewank, Rastrigin, Rosenbrock, Schwefel, Sphere, Weierstrass
    """
    is_realworld = False
    intro = """
        This module implements a multi-task optimization benchmark derived from a single base function,
        with each task being transformed using different rotation matrices. The tasks share the same
        dimensionality and search space bounds, but differ in their problem landscapes due to rotations.
        The number of tasks is fixed at 10. Each task is a transformed version of the specified source problem.
    """
    notes = """
        All tasks are derived from the same base function.
        Tasks are differentiated by applying distinct rotation matrices.
        No shift transformation is applied.
    """
    references = [
        """
        Wang, X., & Jin, Y. (2024). Distilling Ensemble Surrogates for Federated Data-Driven
        Many-Task Optimization. IEEE Transactions on Evolutionary Computation, 1–1.
        https://doi.org/10.1109/TEVC.2024.3428701
        """
    ]

    def __init__(self, dim=10, src_problem: T_SrcProblem = 'Ackley', **kwargs):
        if not isinstance(src_problem, str):
            raise TypeError(f"original_problem should be str, but {type(src_problem)} is given")
        try:
            src_prob_cls = getattr(benchmarks, src_problem)
        except AttributeError:
            raise ValueError(f"{src_problem} is not exist, supported names "
                             f"[Griewank, Rastrigin, Ackley, Schwefel, Sphere"
                             f", Rosenbrock, Weierstrass, Ellipsoid]")
        super().__init__(dim, src_prob_cls, **kwargs)

    def get_info(self):
        task = self._problem[0]
        return {
            "TaskSrc": [task.name],
            "DecDim": [task.dim],
            "Lower": [task.lb[0]],
            "Upper": [task.ub[0]]
        }

    def _init_tasks(self, dim: int, src_prob_cls: Type[SingleTaskProblem], **kwargs):
        funcs = [src_prob_cls(dim, **kwargs) for _ in range(10)]
        datasets = Path(__file__).parents[1] / 'datasets' / 'tevc2024' / 'composition_func_M_D10.mat'
        rot_mats = loadmat(str(datasets))
        mats = [rot_mats[f"M{i + 1}"][0:dim, 0:dim] for i in range(10)]
        for f, mat in zip(funcs, mats):
            f.set_transform(rotation=mat, shift=None)
        return funcs



import subprocess
import sys
import time
from concurrent.futures import ThreadPoolExecutor
from contextlib import contextmanager

from rich import box, progress as rpg
from rich.console import Group, Console
from rich.live import Live
from rich.measure import measure_renderables
from rich.table import Table
from setproctitle import setproctitle
from tabulate import tabulate

from pyfmto.utilities import (
    logger,
    titled_tabulate,
    terminate_popen,
    tabulate_formats as tf, clear_console,
)
from .utils import RunSolutions
from ..framework import Client
from ..utilities.loaders import ExperimentConfig, ConfigLoader

__all__ = ['Launcher']


class Launcher:
    exp_idx: int
    exp: ExperimentConfig
    clients: list[Client]
    table: Table
    progress: rpg.Progress

    def __init__(self, conf_file: str = 'config.yaml'):
        self.conf = ConfigLoader(conf_file).launcher

        # Runtime data
        self._repeat_id = 0
        self._results: list[Client] = []

    def run(self):
        self._setup()
        for self.exp_idx, self.exp in enumerate(self.conf.experiments):
            logger.info(f"\n{self.exp}")
            if self.conf.save:
                self.exp.init_root()
                self.exp.backup_params()
            self._repeating()
        self._remove_bars()
        self.conf.show_summary()

    def _setup(self):
        self.progress = rpg.Progress(
            rpg.TextColumn("[progress.description]{task.description}"),
            rpg.BarColumn(bar_width=None),
            rpg.TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            rpg.TimeRemainingColumn(),
            expand=True,
        )
        setproctitle("AlgClients")

    def _repeating(self):
        self._repeat_id = 0
        self._update_repeat_id()
        while not self._finished:
            self._init_clients()
            self._launch_exp()
            self._save_results()
            self._save_rounds_info()
            self._update_repeat_id()
            time.sleep(1)
        self.exp.success = True

    def _init_clients(self):
        problem = self.exp.problem.initialize()
        clt_params = self.exp.algorithm.params.get('client', {})
        self.clients = [self.exp.algorithm.client(p, **clt_params) for p in problem]

    def _launch_exp(self):
        clear_console()
        with Live(self._panel, console=Console(width=self.width), refresh_per_second=10):
            with self._running_server():
                self._start_clients()

    @property
    def _panel(self) -> Group:
        self._remove_bars()
        self._update_table()
        return Group(self.table, self.progress)

    @property
    def width(self):
        csl = Console(width=1000)
        min_width, max_width = measure_renderables(csl, csl.options, [self.table])
        return min_width

    def _remove_bars(self):
        for bar in self.progress.task_ids:
            self.progress.remove_task(bar)

    def _update_table(self):
        curr_rep = self.conf.repeat * self.exp_idx + self._repeat_id
        total_rep = self.conf.total_repeat
        tab_dict = {
            'running': f"{self.exp_idx+1}/{self.conf.n_exp}",
            'repeating': f"{self._repeat_id}/{self.conf.repeat}",
            'progress': f"[{curr_rep}/{total_rep}][{100 * curr_rep / total_rep:.2f}%]",
            'algorithm': self.exp.algorithm.name,
            'problem': self.exp.problem.name,
            'NPD': self.clients[0].problem.npd,
            'clients': len(self.clients),
            'save': self.conf.save
        }

        def mapper(v):
            if v is True:
                return '[green]yes[/green]'
            elif v is False:
                return '[red]no[/red]'
            elif isinstance(v, int):
                return f'[magenta]{v}[/magenta]'
            else:
                return str(v)

        keys = list(tab_dict.keys())
        values = list(tab_dict.values())
        colored_values = list(map(mapper, values))

        tab_csl = Table(box=box.ROUNDED)
        [tab_csl.add_column(k, justify='center') for k in keys]
        tab_csl.add_row(*colored_values)
        self.table = tab_csl

        alignment = ['center'] * len(keys)
        original_tab = {k: [v] for k, v in zip(keys, values)}
        tab_log = tabulate(original_tab, headers='keys', tablefmt='rounded_grid', colalign=alignment)
        logger.info(f"\n{tab_log}")

    @contextmanager
    def _running_server(self):
        server = self.exp.algorithm.server
        kwargs = self.exp.algorithm.params.get('server', {})

        cmd = [
            sys.executable, "-c",
            "from pyfmto.utilities import logger; "
            f"from {server.__module__} import {server.__name__}; "
            f"logger.setLevel('{self.conf.loglevel}'); "
            f"srv = {server.__name__}(**{repr(kwargs)}); "
            f"srv.start()"
        ]

        process = subprocess.Popen(
            cmd,
            stdin=subprocess.DEVNULL,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE
        )
        logger.info("Server started.")
        time.sleep(3)
        try:
            yield process
        finally:
            terminate_popen(process)
            logger.debug("Server terminated.")

    def _start_clients(self):
        n_clients = len(self.clients)
        for c in self.clients:
            bar = self.progress.add_task(c.name, total=c.fe_max)
            c.progress = self.progress
            c.bar = bar
        pool = ThreadPoolExecutor(max_workers=n_clients)
        futures = [pool.submit(c.start) for c in self.clients]
        pool.shutdown(wait=True)
        self._results = [fut.result() for fut in futures]

    def _save_rounds_info(self):
        if self.conf.backup:
            tables = {clt.id: clt for clt in self._results}
            info_str = ''
            for cid in sorted(tables.keys()):
                clt = tables[cid]
                tab = titled_tabulate(
                    f"{clt.name} {clt.problem.name}({clt.dim}D)",
                    '=', clt.rounds_info, headers='keys', tablefmt=tf.rounded_grid
                )
                info_str = f"{info_str}{tab}\n"
            res_file = self.exp.result_name(self._repeat_id)
            log_dir = res_file.with_name('rounds_info')
            log_name = res_file.with_suffix('.log').name
            log_dir.mkdir(parents=True, exist_ok=True)
            with open(log_dir / log_name, 'w') as f:
                f.write(info_str)

    def _save_results(self):
        if self.conf.save:
            run_solutions = RunSolutions()
            for clt in self._results:
                run_solutions.update(clt.id, clt.solutions)
            run_solutions.to_msgpack(self.exp.result_name(self._repeat_id))

    def _update_repeat_id(self):
        if self.conf.save:
            self._repeat_id = self.exp.num_results + 1
        else:
            self._repeat_id += 1

    @property
    def _finished(self):
        return self._repeat_id > self.conf.repeat



import importlib
from pyfmto.problems import SingleTaskProblem as Stp, MultiTaskProblem as Mtp

__all__ = ['Cec2022']


class _BenchmarksCec2022(Stp):
    def __init__(self, fid, dim: int, **kwargs):
        cec2022 = importlib.import_module('opfunu.cec_based.cec2022')
        func = getattr(cec2022, f"F{fid}2022")(dim)
        super().__init__(dim, 1, lb=func.lb, ub=func.ub, **kwargs)
        self.func = func
        self.set_x_global(func.x_global)

    @property
    def name(self) -> str:
        return f"CEC2022F{self.id}"

    def _eval_single(self, x):
        return self.func.evaluate(x)


class Cec2022(Mtp):
    """
    dim: 10  # 10 or 20
    """
    is_realworld = False
    intro = """
        Synthetic of CEC 2022 benchmark problems
    """

    def __init__(self, dim=10, **kwargs):
        if dim not in [10, 20]:
            raise ValueError('CEC2022 only support 10D and 20D')
        super().__init__(dim, **kwargs)

    def _init_tasks(self, dim, **kwargs):
        funcs = [_BenchmarksCec2022(fid, dim, **kwargs) for fid in range(1, 13)]
        return funcs



import copy
import importlib
import inspect
import os
import textwrap
from deepdiff import DeepDiff
from importlib import import_module
from itertools import product
from pathlib import Path
from pydantic import BaseModel, ConfigDict
from rich import box
from rich.console import Console
from rich.table import Table
from ruamel.yaml import CommentedMap
from textwrap import indent
from typing import Type, Any, Union, Literal

from pyfmto.framework.client import Client
from pyfmto.framework.server import Server
from pyfmto.problems import MultiTaskProblem, realworld, synthetic
from . import titled_tabulate, clear_console
from .io import parse_yaml, dumps_yaml, load_yaml, save_yaml
from .loggers import logger

__all__ = [
    'list_algorithms',
    'load_algorithm',
    'list_problems',
    'load_problem',
    'init_problem',
    'ProblemData',
    'ConfigLoader',
    'AlgorithmData',
    'ExperimentConfig',
    'ReporterConfig',
    'LauncherConfig',
]


def recursive_to_pure_dict(data: Any) -> dict[str, Any]:
    """
    Recursively convert nested dict and CommentedMap objects to a pure Python
    dictionary to avoid YAML serialization issues.
    """
    if isinstance(data, (dict, CommentedMap)):
        for k, v in data.items():
            data[k] = recursive_to_pure_dict(v)
    else:
        return data
    return data


def combine_params(params: dict[str, Union[Any, list[Any]]]) -> list[dict[str, Any]]:
    values = []
    for key, value in params.items():
        if isinstance(value, list):
            values.append(value)
        else:
            values.append([value])
    result = []
    for combination in product(*values):
        result.append(dict(zip(params.keys(), combination)))
    return result


def list_algorithms(print_it=False):
    alg_dir = Path().cwd() / 'algorithms'
    if alg_dir.exists():
        folders = os.listdir(alg_dir)
        alg_names = [alg_name for alg_name in folders if alg_name.isupper()]
    else:
        logger.error(f"'algorithms' folder not found in {alg_dir.parent}.")
        alg_names = []

    algorithms: dict[str, AlgorithmData] = {}
    for name in alg_names:
        try:
            algorithms[name] = load_algorithm(name)
        except Exception as e:
            logger.error(f"Failed to load {name}: {e}")

    if print_it:
        if alg_dir.exists():
            alg_str = '\n'.join(list(algorithms.keys()))
            print(f"Found {len(algorithms)} available algorithms: \n{textwrap.indent(alg_str, ' ' * 4)}")
        else:
            print(f"'algorithms' folder not found in {alg_dir.parent}.")
    return algorithms


def load_algorithm(name: str) -> 'AlgorithmData':
    name = name.upper()
    alg_dir = Path().cwd() / 'algorithms' / name
    if alg_dir.exists():
        module = importlib.import_module(f"algorithms.{name}")
    else:
        if alg_dir.parent.exists():
            raise ValueError(f"Algorithm {name} not found.")
        else:
            raise FileNotFoundError(f"'algorithms' folder not found in {alg_dir.parent.parent}.")

    clt, srv = None, None
    for attr_name in dir(module):
        attr = getattr(module, attr_name)
        if inspect.isclass(attr):
            if issubclass(attr, Client):
                clt = attr
            if issubclass(attr, Server):
                srv = attr
        if clt and srv:
            return AlgorithmData(name, clt, srv)

    msg: list[str] = [f'Load algorithm {name} failed:']
    if not clt:
        msg.append("  Client not found.")
    if not srv:
        msg.append("  Server not found.")
    raise ModuleNotFoundError('\n'.join(msg))


def init_problem(name: str, **kwargs) -> MultiTaskProblem:
    prob = load_problem(name)
    prob.params_update = kwargs
    return prob.initialize()


def load_problem(name: str) -> 'ProblemData':
    problems = list_problems()
    if name in problems.keys():
        return problems[name]
    else:
        raise ValueError(f"Problem '{name}' not found, use 'pyfmto show problems' to see available problems.")


def list_problems(print_it=False) -> dict[str, 'ProblemData']:
    problems: dict[str, ProblemData] = {}
    for module in [realworld, synthetic]:
        problems.update(collect_problems(module))
    prob_dir = Path().cwd() / 'problems'
    if prob_dir.exists():
        problems.update(collect_problems(import_module('problems')))
    if print_it:
        print(f"Found {len(problems)} available problems:")
        print(textwrap.indent('\n'.join(list(problems.keys())), ' ' * 4))
    return problems


def collect_problems(module) -> dict[str, 'ProblemData']:
    problems = {}
    for name in dir(module):
        cls = getattr(module, name)
        if inspect.isclass(cls) and issubclass(cls, MultiTaskProblem) and cls != MultiTaskProblem:
            problems[name] = ProblemData(cls)
    return problems


class AlgorithmData:

    def __init__(self, name: str, client: Any = None, server: Any = None):
        self.name_orig = name
        self.name_alias = ''
        self.client: Type[Client] = client
        self.server: Type[Server] = server
        self.params_default: dict[str, dict[str, Any]] = {}
        self.params_update: dict[str, dict[str, Any]] = {}
        self.__parse_default_params()

    def __parse_default_params(self):
        c_doc = self.client.__doc__ if self.client else None
        s_doc = self.server.__doc__ if self.server else None
        c_args = parse_yaml(c_doc) if c_doc else {}
        s_args = parse_yaml(s_doc) if s_doc else {}
        if c_args:
            self.params_default.update({'client': c_args})
        if s_args:
            self.params_default.update({'server': s_args})

    def copy(self) -> 'AlgorithmData':
        return copy.deepcopy(self)

    @property
    def params(self) -> dict[str, dict[str, Any]]:
        kwargs = copy.deepcopy(self.params_default)
        for k in ['client', 'server']:
            for k2, v2 in self.params_update.get(k, {}).items():
                if k not in kwargs:
                    kwargs[k] = {}
                kwargs[k][k2] = v2
        return kwargs

    @property
    def params_diff(self) -> str:
        return DeepDiff(self.params_default, self.params).pretty()

    @property
    def name(self) -> str:
        return self.name_orig if not self.name_alias else self.name_alias

    @property
    def params_yaml(self) -> str:
        if self.params_default:
            return dumps_yaml(self.params_default)
        else:
            return f"Algorithm '{self.name}' no configurable parameters."


class ProblemData:

    def __init__(self, problem: Type[MultiTaskProblem]):
        self.problem: Type[MultiTaskProblem] = problem
        self.params_default: dict[str, Any] = {
            'npd': 1,
            'random_ctrl': 'weak',
            'seed': 123,
        }
        self.params_update: dict[str, Any] = {}
        self.__parse_default_params()

    def __parse_default_params(self):
        p_doc = self.problem.__doc__
        self.params_default.update(parse_yaml(p_doc))

    def copy(self) -> 'ProblemData':
        return copy.deepcopy(self)

    @property
    def params(self) -> dict[str, Any]:
        params = copy.deepcopy(self.params_default)
        params.update(self.params_update)
        dim = params.get('dim', 0)
        if dim > 0:
            if 'fe_init' not in params:
                params.update(fe_init=5 * dim)
            if 'fe_max' not in params:
                params.update(fe_max=11 * dim)
        return params

    @property
    def params_diff(self) -> str:
        return DeepDiff(self.params_default, self.params).pretty()

    def initialize(self) -> MultiTaskProblem:
        return self.problem(**self.params)

    @property
    def name(self) -> str:
        base_name = self.problem.__name__
        return f"{base_name}_{self.dim_str}" if self.dim > 0 else base_name

    @property
    def npd(self) -> int:
        return self.params.get('npd', 0)

    @property
    def npd_str(self) -> str:
        return f"NPD{self.npd}" if self.npd > 0 else ""

    @property
    def dim(self) -> int:
        return self.params.get('dim', 0)

    @property
    def dim_str(self) -> str:
        if self.dim > 0:
            return f"{self.dim}D"
        else:
            return ""

    @property
    def params_yaml(self) -> str:
        if self.params_default:
            return dumps_yaml(self.params_default)
        else:
            return f"Problem '{self.name}' no configurable parameters."


class ExperimentConfig:

    def __init__(
            self,
            algorithm: AlgorithmData,
            problem: ProblemData,
            root: str
    ):
        self.algorithm = algorithm
        self.problem = problem
        self._root = Path(root)
        self.success = False

    @property
    def root(self) -> Path:
        return self._root / self.algorithm.name / self.problem.name / self.problem.npd_str

    @property
    def prefix(self) -> str:
        fe_init = self.problem.params.get('fe_init')
        fe_max = self.problem.params.get('fe_max')
        seed = self.problem.params.get('seed')
        fe_i = "" if fe_init is None else f"FEi{fe_init}_"
        fe_m = "" if fe_max is None else f"FEm{fe_max}_"
        seed = "" if seed is None else f"Seed{seed}_"
        return f"{fe_i}{fe_m}{seed}"

    def result_name(self, file_id: int):
        return self.root / f"{self.prefix}Rep{file_id:02d}.msgpack"

    @property
    def num_results(self) -> int:
        if not self.root.exists():
            return 0
        prefix = self.result_name(0).name.split('Rep')[0]
        suffix = '.msgpack'
        results = [f for f in os.listdir(self.root) if f.startswith(prefix) and f.endswith(suffix)]
        return len(results)

    @property
    def params_dict(self) -> dict[str, Any]:
        data = {
            'algorithm': {
                'name': self.algorithm.name,
                'params': self.algorithm.params,
                'default': self.algorithm.params_default,
                'update': self.algorithm.params_update,
            },
            'problem': {
                'name': self.problem.name,
                'params': self.problem.params,
                'default': self.problem.params_default,
                'update': self.problem.params_update,
            }
        }
        return recursive_to_pure_dict(data)

    def backup_params(self):
        self.root.parent.mkdir(parents=True, exist_ok=True)
        save_yaml(self.params_dict, self.root.parent / "parameters.yaml")

    def init_root(self):
        self.root.mkdir(parents=True, exist_ok=True)

    def __str__(self):
        tab = {
            'Algorithm': [self.algorithm.name],
            'Problem': [self.problem.name],
            'NPD': [self.problem.npd_str],
            'Dimension': [self.problem.dim_str],
        }
        return titled_tabulate("Experiment", '=', tab, tablefmt='rounded_grid')

    def __repr__(self):
        info = [
            f"Alg({self.algorithm.name})",
            f"Prob({self.problem.name})",
            f"NPD({self.problem.params['npd']})",
            f"Dim({self.problem.params.get('dim', '-')})",
        ]

        return ' '.join(info)


class LauncherConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    results: str
    repeat: int
    seed: int
    backup: bool
    save: bool
    loglevel: str
    algorithms: list[str]
    problems: list[str]
    experiments: list[ExperimentConfig] = []

    def show_summary(self):
        tab = Table(
            title="Experiments Summary",
            title_justify="center",
            box=box.ROUNDED,
        )
        tab.add_column('Algorithm', justify='center', style="cyan")
        tab.add_column('Original', justify='center', style="cyan")
        tab.add_column('Problem', justify='center', style="magenta")
        tab.add_column('NPD', justify='center', style="yellow")
        tab.add_column('Success', justify='center')

        for exp in self.experiments:
            tab.add_row(
                exp.algorithm.name,
                exp.algorithm.name_orig,
                exp.problem.name,
                exp.problem.npd_str,
                '[green]yes[/green]' if exp.success else '[red]no[/red]'
            )
        clear_console()
        Console().print(tab)

    @property
    def n_exp(self) -> int:
        return len(self.experiments)

    @property
    def total_repeat(self) -> int:
        return self.n_exp * self.repeat


class ReporterConfig(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)
    results: str
    algorithms: list[list[str]]
    problems: list[str]
    formats: list[str]
    params: dict[str, Any] = {}
    experiments: list[ExperimentConfig] = []
    groups: list[tuple[list[str], str, str]] = []

    @property
    def root(self) -> Path:
        return Path(self.results)


class ConfigLoader:
    """
    launcher:
        results: out/results  # [optional] save results to this directory
        repeat: 2             # [optional] repeat each experiment for this number of times
        seed: 123             # [optional] random seed
        save: true            # [optional] save results to disk
        backup: false          # [optional] backup parameters to disk
        loglevel: INFO        # [optional] log level [CRITICAL, ERROR, WARNING, INFO, DEBUG], default INFO
        algorithms: []        # run these algorithms
        problems: []          # run each algorithm on these problems
    reporter:
        formats: [excel]      # [optional] generate these reports
    """

    def __init__(self, config: str = 'config.yaml'):
        self.config_default = parse_yaml(self.__class__.__doc__)
        self.config_update = load_yaml(config)
        self.config = copy.deepcopy(self.config_default)

        self.merge_global_config_from_updates()
        self.fill_launcher_config_from_reporter()

    def merge_global_config_from_updates(self):
        for key, value in self.config_update.items():
            if key in self.config:
                self.config[key].update(value)
            else:
                self.config[key] = value

    def fill_launcher_config_from_reporter(self) -> None:
        launcher_params = self.config['launcher']
        for key in ['results', 'problems', 'algorithms']:
            if key in self.config['reporter']:
                continue
            if key == 'algorithms':
                self.config['reporter'][key] = [launcher_params[key]]
            else:
                self.config['reporter'][key] = launcher_params[key]

    @property
    def launcher(self) -> LauncherConfig:
        self.check_config_issues('launcher')
        conf = LauncherConfig(**self.config['launcher'])
        logger.setLevel(conf.loglevel)
        algorithms = self.gen_alg_list(conf.algorithms)
        problems = self.gen_prob_list(conf.problems)
        logger.debug(f"algorithms: {[alg.name for alg in algorithms]}")
        logger.debug(f"problems: {[prob.name for prob in problems]}")
        conf.experiments = [ExperimentConfig(alg, prob, conf.results) for alg, prob in product(algorithms, problems)]
        return conf

    @property
    def reporter(self) -> ReporterConfig:
        self.check_config_issues('reporter')
        conf = ReporterConfig(**self.config['reporter'])
        alg_names = list(set(sum(conf.algorithms, [])))
        algorithms = [AlgorithmData(name) for name in alg_names]
        problems = self.gen_prob_list(conf.problems)
        conf.experiments = [
            ExperimentConfig(alg, prob, conf.results)
            for alg, prob in product(algorithms, problems)
        ]
        conf.groups = [
            (algs, prob.name, prob.npd_str)
            for algs, prob in list(product(conf.algorithms, problems))
        ]
        return conf

    def gen_alg_list(self, names: list[str]) -> list[AlgorithmData]:
        algorithms: list[AlgorithmData] = []
        available_algs = list_algorithms()
        for name_alias in names:
            alg_params = self.config.get('algorithms', {}).get(name_alias, {})
            alg_name = alg_params.pop('base', name_alias)
            if alg_name not in available_algs:
                logger.error(f"Algorithm {alg_name} is not available.")
                continue
            alg_data = available_algs[alg_name].copy()
            alg_data.name_alias = name_alias
            alg_data.params_update = alg_params
            algorithms.append(alg_data)
        return algorithms

    def gen_prob_list(self, names: list[str]) -> list[ProblemData]:
        available_probs = list_problems()
        problems: list[ProblemData] = []
        for n in names:
            if n not in available_probs:
                logger.error(f"Problem {n} is not available.")
                continue
            prob_params = self.config.get('problems', {}).get(n, {})
            params_variations = combine_params(prob_params)
            for params in params_variations:
                prob_data = available_probs[n].copy()
                prob_data.params_update = params
                problems.append(prob_data)
        return problems

    def check_config_issues(self, name: Literal['launcher', 'reporter']) -> None:
        if name == 'launcher':
            issues = self.check_launcher_config()
        else:
            issues = self.check_reporter_config()
        if issues:
            detail = indent('\n'.join(issues), ' ' * 4)
            msg = f"{name.title()} configuration issues:\n{detail}"
            logger.error(msg)
            raise ValueError(msg)

    def check_launcher_config(self) -> list[str]:
        issues = []
        launcher = self.config['launcher']
        loglevel = ['CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG']
        if not launcher.get('results'):
            issues.append("No results directory specified in launcher.")
        if launcher.get('repeat') <= 0:
            issues.append("Invalid repeat number specified in launcher. Must be greater than 0.")
        if not isinstance(launcher.get('save'), bool):
            issues.append("Invalid save option specified in launcher. Must be True or False.")
        if launcher.get('loglevel') not in loglevel:
            issues.append(f"Invalid log level specified in launcher. Choices: {loglevel}")
        if not launcher.get('algorithms'):
            issues.append("No algorithms specified in launcher.")
        if not launcher.get('problems'):
            issues.append("No problems specified in launcher.")
        return issues

    def check_reporter_config(self) -> list[str]:
        issues = []
        reporter = self.config['reporter']
        if not reporter.get('results'):
            issues.append("No results directory specified in reporter or launcher.")
        if not reporter.get('algorithms', []):
            issues.append("No algorithms specified in reporter or launcher.")
        else:
            validate_values: list[list[str]] = []
            for item in reporter['algorithms']:
                if isinstance(item, str) and item:
                    validate_values.append([item])
                elif isinstance(item, list) and item:
                    validate_values.append(item)
                else:
                    issues.append(f"Invalid value [type:{type(item)}, value:{item}] specified in reporter.")
                reporter['algorithms'] = validate_values
        return issues



import argparse
import sys
from pathlib import Path

from .tools import matched_str_head
from ..experiments import Launcher, Reports, list_report_formats, show_default_conf
from .loaders import list_problems, load_algorithm, list_algorithms, load_problem


def update_path():
    current_dir = Path().cwd()
    if str(current_dir) not in sys.path:
        sys.path.append(str(current_dir))


def main():
    update_path()
    parser = argparse.ArgumentParser(
        description='PyFMTO: Python Library for Federated Many-task Optimization Research'
    )
    subparsers = parser.add_subparsers(dest='command', help='Available commands')

    # Run command
    run_parser = subparsers.add_parser('run', help='Run experiments')
    run_parser.add_argument(
        '-c', '--config', type=str, default='config.yaml',
        help='Path to configuration file (default: config.yaml)')

    # Report command
    report_parser = subparsers.add_parser('report', help='Generate reports')
    report_parser.add_argument(
        '-c', '--config', type=str, default='config.yaml',
        help='Path to configuration file (default: config.yaml)')

    # List command
    list_parser = subparsers.add_parser('list', help='List available options')
    list_parser.add_argument(
        'name', type=str, help='Name of the option to list'
    )

    # Show command
    show_parser = subparsers.add_parser('show', help='Show default configurations')
    show_parser.add_argument(
        'name', type=str,
        help="Name of the configuration to show, any things that can be list by the 'list' command"
    )

    args = parser.parse_args()

    if args.command == 'run':
        launcher = Launcher(conf_file=args.config)
        launcher.run()
    elif args.command == 'report':
        reports = Reports(conf_file=args.config)
        reports.generate()
    elif args.command == 'list':
        full_name = matched_str_head(args.name, ['problems', 'algorithms', 'reports'])
        if full_name == 'problems':
            list_problems(print_it=True)
        elif full_name == 'algorithms':
            list_algorithms(print_it=True)
        elif full_name == 'reports':
            list_report_formats(print_it=True)
    elif args.command == 'show':
        t, v = args.name.split('.')
        full_name = matched_str_head(t, ['problems', 'algorithms', 'reports'])
        if full_name == 'problems':
            prob = load_problem(v)
            print(prob.params_yaml)
        elif full_name == 'algorithms':
            alg = load_algorithm(v)
            print(alg.params_yaml)
        elif full_name == 'reports':
            show_default_conf(v)
        else:
            print(f"No matched group for {t}.")



import copy
import numpy as np
import pandas as pd
import seaborn as sns
import textwrap
from abc import ABC, abstractmethod
from matplotlib import pyplot as plt
from matplotlib.ticker import FuncFormatter
from numpy import ndarray
from pathlib import Path
from pyDOE import lhs
from scipy.stats import kendalltau, spearmanr, pearsonr
from tabulate import tabulate
from typing import Union, Optional, Literal

from pyfmto.utilities import StrColors, Cmaps, SeabornStyles
from pyfmto.utilities.schemas import T_Bound, STPConfig, TransformerConfig, FunctionInputs, PlottingArgs
from .solution import Solution

__all__ = [
    'Transformer',
    'MultiTaskProblem',
    'SingleTaskProblem',
]


class Transformer:
    def __init__(self, dim: int):
        self._transform = TransformerConfig(dim=dim)

    @property
    def dim(self):
        return self._transform.dim

    @property
    def shift(self):
        return self._transform.shift

    @property
    def rotation(self):
        return self._transform.rotation

    @property
    def rotation_inv(self):
        return self._transform.rotation_inv

    def set_transform(self, rotation: Optional[ndarray] = None, shift: Optional[ndarray] = None):
        self._transform = TransformerConfig(
            dim=self.dim,
            rotation=rotation,
            shift=shift
        )

    def transform_x(self, x):
        return (x - self.shift) @ self.rotation

    def inverse_transform_x(self, x):
        return (x @ self.rotation_inv) + self.shift


class SingleTaskProblem(ABC):
    _id: int
    _x_global: ndarray
    _partition: ndarray
    _solutions: Solution
    _transformer: Transformer
    auto_update_solutions = False

    def __init__(self, dim: int, obj: int, lb: T_Bound, ub: T_Bound, **kwargs):
        """
        Initialize a `SingleTaskProblem` instance.

        Parameters
        ----------
        dim : int
            Dimension of the decision space (number of input variables).
        obj : int
            Dimension of the objective space (number of output objectives).
        lb : T_Bound
            Lower bounds for the decision variables. Can be a scalar or an array-like of shape (dim,).
        ub : T_Bound
            Upper bounds for the decision variables. Must have the same shape as [x_lb].
        **kwargs : dict, optional
            Additional keyword arguments. Supported options are:

            - fe_init : int, optional
              Initial function evaluations.
            - fe_max : int, optional
              Maximum function evaluation budget.
            - npd : int, default=1
              Number of partitions per dimension for non-IID simulation.

        Raises
        ------
        ValueError
            If invalid types or values are provided for `dim`, `obj`, `x_lb`, or `x_ub`.
        TypeError
            If unsupported argument types are passed for `x_lb` or `x_ub`.
        ValueError
            If any unrecognized keyword arguments are provided.

        Notes
        -----
        - The actual bounds are stored in `self.x_lb` and `self.x_ub`.
        - A partition can be generated if `npd > 1` to simulate non-IID scenarios.
        - The optimal solutions and Pareto Front (`self.optimum`, `self.PF`) are precomputed with default settings.
        """
        self._config = STPConfig(dim=dim, obj=obj, lb=lb, ub=ub, **kwargs)
        self._x_global = np.zeros(self.dim)
        self._transformer = Transformer(self.dim)

    def __str__(self):
        lb, ub = self.lb, self.ub
        if np.all(self.lb == self.lb[0]) and np.all(self.ub == self.ub[0]):
            lb, ub = self.lb[0], self.ub[0]
        func_info = {
            "ID": [self.id],
            "OriFunc": [self.name],
            "DecDim": [self.dim],
            "lb": [lb],
            "ub": [ub]
        }

        tab = tabulate(func_info, headers="keys", missingval='-', tablefmt="rounded_grid")
        return tab

    def __repr__(self):
        return (f"{type(self).__name__}("
                f"ID={self.id}, "
                f"dim={self.dim}, obj={self.obj}, "
                f"lb={self.lb}, ub={self.ub}, "
                f"init_fe={self.fe_init}, max_fe={self.fe_max})")

    def init_partition(self):
        """
        Set attribute 'partition' according to 'npd' to simulate the non-IID settings.

        Notes
        -----
        It would do nothing if npd=1.

        - >>> Example of the bounds chosen process when dim=5 and npd=5
        - [[0.0  0.2  0.4  0.6  0.8->1.0],
        -  [0.0  0.2->0.4  0.6  0.8  1.0],
        -  [0.0  0.2->0.4  0.6  0.8  1.0],
        -  [0.0  0.2  0.4->0.6  0.8  1.0],
        -  [0.0  0.2  0.4  0.6->0.8  1.0]]
        - >>> partition will be
        - [[0.8 0.2 0.2 0.4 0.6]   <- lb
        -  [1.0 0.4 0.4 0.6 0.8]]  <- ub
        """
        p_mat = [np.linspace(0, 1, self.npd + 1) for _ in range(self.dim)]
        p_sampled = np.random.randint(0, self.npd, size=self.dim)
        lb = []
        ub = []
        for idx, p in zip(p_sampled, p_mat):
            lb.append(p[idx])
            ub.append(p[idx + 1])
        partition = np.array([lb, ub]) * (self.ub - self.lb) + self.lb
        self._partition = partition

    def gen_plot_data(
            self,
            dims: tuple[int, int] = (0, 1),
            n_points: int = 100,
            fixed=None,
            scale_mode: Literal['', 'y', 'xy'] = '',
    ) -> tuple[ndarray, ndarray, ndarray, PlottingArgs]:
        """

        Parameters
        ----------
        dims:
            Selected dimensions to calculate y.
        n_points:
            Number of points in one dimension, total use n_points**2 points.
        fixed:
            Fixed value of no-selected dimensions
        scale_mode:
            If pass 'xy', normalize both the decision space and objective space to (0, 1).
            If pass 'y', scale objective value to match the decision space, which will keep
            the shift information.
            Any other value will be ignored.
        Returns
        -------
        Tuple:
            The grid data and validated args ``(D1,D2,Z,args)``.
        """
        args = PlottingArgs(
            dim=self.dim,
            ub=self.ub,
            lb=self.lb,
            dims=dims,
            n_points=n_points,
            fixed=fixed,
        )
        dim1, dim2 = args.dims
        lb1, ub1 = self.lb[dim1], self.ub[dim1]
        lb2, ub2 = self.lb[dim2], self.ub[dim2]
        d1 = np.linspace(lb1, ub1, n_points)
        d2 = np.linspace(lb2, ub2, n_points)
        D1, D2 = np.meshgrid(d1, d2)
        points: ndarray = np.ones(shape=(n_points, n_points, self.dim)) * args.fixed
        points[:, :, dim1] = D1
        points[:, :, dim2] = D2
        Z = np.apply_along_axis(self.evaluate, axis=-1, arr=points)
        Z = Z.squeeze()
        if scale_mode == 'xy':
            dn = np.linspace(0, 1, n_points)
            D1, D2 = np.meshgrid(dn, dn)
            Z = (Z - np.min(Z)) / (np.max(Z) - np.min(Z) + 1e-8)  # type: ignore
        elif scale_mode == 'y':
            x_range = self.ub[0] - self.lb[0]
            Z = x_range * (Z - np.min(Z)) / (np.max(Z) - np.min(Z) + 1e-8)  # type: ignore
        else:
            pass
        return D1, D2, Z, args

    def plot_2d(
            self,
            filename: Optional[Union[str, Path]] = None,
            dims: tuple[int, int] = (0, 1),
            n_points: int = 100,
            figsize: tuple[float, float, float] = (5., 4., 1.),
            cmap: Union[str, Cmaps] = Cmaps.viridis,
            levels: int = 30,
            labels: tuple[Optional[str], Optional[str]] = (None, None),
            title: Optional[str] = None,
            alpha: float = 0.7,
            fixed: Union[float, ndarray, None] = None):
        """
        Plot the function's 2D contour.

        Parameters
        ----------
        filename:
            Filename, show image if pass None
        dims : list, tuple
            The selected dimensions indices to be drawn.
        n_points : int
            Using total n_points**2 points to draw the contour
        figsize : tuple
            figsize (width, height, scale)
        cmap : str
            The cmap of matplotlib.pyplot.contourf function. The `pyfmto.utilities.Cmaps` class
            can help you try different options easier.
        levels : int
            The levels parameter of contourf function
        labels : tuple
            The figure x, y labels.
        title :
            The figure title
        alpha : float
            The alpha parameter of contourf function
        fixed : float, ndarray
            Fixed values for all the unused dimensions, if pass a ndarray, its shape should be (dim, ).
        """
        w, h, s = figsize
        plt.figure(figsize=(w * s, h * s))
        D1, D2, Z, args = self.gen_plot_data(dims=dims, n_points=n_points, fixed=fixed)
        cont = plt.contourf(D1, D2, Z, levels=levels, cmap=str(cmap), alpha=alpha)
        cbar = plt.colorbar(cont)
        cbar.set_label('Function Value')
        cbar.formatter = FuncFormatter(lambda x, pos: f'{x:.2e}' if abs(x) > 1e4 else f'{x:.2f}')
        cbar.update_ticks()
        xl, yl = labels
        xl = f'X{args.dims[0]}' if xl is None else xl
        yl = f'X{args.dims[1]}' if yl is None else yl
        _title = f'T{self.id} {self.name}' if title is None else title
        plt.xlabel(xl)
        plt.ylabel(yl, rotation=0)
        plt.title(_title)
        plt.tight_layout()
        if filename is not None:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()

    def plot_3d(
            self,
            filename: Optional[Union[str, Path]] = None,
            dims: tuple[int, int] = (0, 1),
            n_points: int = 100,
            figsize: tuple[float, float, float] = (5., 4., 1.),
            cmap: Union[str, Cmaps] = Cmaps.viridis,
            levels: int = 30,
            labels: tuple[Optional[str], Optional[str], Optional[str]] = (None, None, None),
            title: Optional[str] = None,
            alpha=0.7,
            fixed=None,
    ):
        """
        Plot the function's 3D surface.

        Parameters
        ----------
        filename:
            Filename, show image if pass None.
        dims : list, tuple
            The selected dimensions indices to be drawn.
        n_points : int
            Using total n_points**2 points to draw the contour.
        figsize : tuple
            figsize (width, height, scale).
        cmap : Union[str, Cmaps]
            The cmap of matplotlib.pyplot.contourf function, use Cmaps for easy selection.
        levels : int
            The levels of the contourf.
        labels : tuple
            The figure x, y, z labels.
        title :
            The figure title.
        alpha : float
            The alpha parameter of contourf function.
        fixed : float, ndarray
            Fixed values for all the unused dimensions, if pass a ndarray, its shape should be (dim,).
        """
        w, h, s = figsize
        D1, D2, Z, args = self.gen_plot_data(dims=dims, n_points=n_points, fixed=fixed)
        fig = plt.figure(figsize=(w * s, h * s))
        ax = fig.add_subplot(111, projection='3d')
        surf = ax.plot_surface(D1, D2, Z, cmap=str(cmap), edgecolor='none', alpha=alpha)
        ax.contour(D1, D2, Z, zdir='z', offset=np.min(Z), cmap=str(cmap), levels=levels)
        cbar = fig.colorbar(surf, ax=ax)
        cbar.set_label('Function Value')
        cbar.formatter = FuncFormatter(lambda x, pos: f'{x:.2e}' if abs(x) > 1e4 else f'{x:.2f}')
        cbar.update_ticks()

        xl, yl, zl = labels
        xl = f'X{args.dims[0]}' if xl is None else xl
        yl = f'X{args.dims[1]}' if yl is None else yl
        zl = 'Y' if zl is None else zl
        _title = f"T{self.id} {self.name}" if title is None else title
        ax.view_init(elev=20, azim=-120)
        ax.set_xlabel(xl)
        ax.set_ylabel(yl)
        ax.set_zlabel(zl)
        ax.set_title(_title)
        plt.tight_layout()
        if filename is not None:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()

    def iplot_3d(
            self,
            dims: tuple[int, int] = (0, 1),
            n_points: int = 100,
            font_size: float = 10,
            cmap: Union[str, Cmaps] = Cmaps.viridis,
            color: Union[str, StrColors, tuple[float, float, float], None] = None,
            show_grid: bool = True,
            scale_mode: Literal['', 'y', 'xy'] = 'y',
            fixed=None,
            plotter=None
    ) -> None:
        """
        Plot the function's 3D surface in an interactive window.

        Parameters
        ----------
        dims : list, tuple
            The selected dimensions indices to be drawn.
        n_points : int
            Using total n_points**2 points to draw the contour
        font_size:
            Fontsize of subplots title.
        cmap : Union[str, Cmaps]
            The cmap of matplotlib.pyplot.contourf function, use Cmaps for easy selection.
        color:
            Use to make each mesh have a single solid color. ``cmap`` will be overridden
            if color is specified. Either a string, RGB list, or hex color string. For
            example: ``color='white'``, ``color=StrColors.white``, ``color='w'``,
            ``color=[1.0,1.0,1.0]``, or ``color='#FFFFFF'``.
        show_grid :
            Show coordinate plane.
        scale_mode:
            if pass 'xy', normalize both the decision space and objective space to (0, 1).
            if pass 'y', scale objective value to match the decision space, which will keep
            the original decision space information.
        fixed : float, ndarray
            Fixed values for all the unused dimensions, if pass a ndarray, its shape should be (dim, ).
        plotter:
            The pyvista.Plotter object. If the plotter is not None, plot on it and return, else plot
            on a new created plotter and show.
        """
        import pyvista as pv
        show = plotter is None
        if plotter is None:
            plotter = pv.Plotter(shape=(1, 1))
            plotter.subplot(0, 0)
        D1, D2, Z, args = self.gen_plot_data(dims=dims, fixed=fixed, n_points=n_points, scale_mode=scale_mode)
        grid = pv.StructuredGrid(D1, D2, Z)
        if color is None:
            plotter.add_mesh(grid, scalars=grid.points[:, 2], cmap=str(cmap))
            plotter.remove_scalar_bar()
        else:
            _color = str(color) if isinstance(color, StrColors) else color
            plotter.add_mesh(grid, color=_color)
        plotter.add_title(f"T{self.id} {self.name}", font_size=font_size)
        if show_grid:
            plotter.show_grid(
                xtitle=f'X{args.dims[0]}',
                ytitle=f'X{args.dims[1]}',
                ztitle='Y',
                font_size=8,
                location='outer'
            )
        plotter.add_axes()
        if show:
            plotter.show()

    def set_x_global(self, x_global: Optional[ndarray]):
        """
        Setting the global optimum solution x.

        Notes
        -----
        Default to zero vector if it doesn't set,
        pass a None if it is unknown.
        """
        if x_global is None:
            self._x_global = np.array([])
        elif isinstance(x_global, ndarray):
            self._x_global = x_global
        else:
            raise ValueError(f"x_global must ndarray or None, got ({x_global}) instead.")

    def set_transform(self, rotation: Optional[ndarray] = None, shift: Optional[ndarray] = None):
        self._transformer.set_transform(rotation, shift)

    def set_id(self, _id: int):
        self._id = _id

    def init_solutions(self):
        """
        Initialize solutions for the problem using either LHS sampling or random sampling within a partition.

        If no partition is defined (`self._partition` is None), Latin Hypercube Sampling (LHS)
        is used to generate initial solutions. Otherwise, random uniform sampling within the
        defined partition is performed.

        The generated solutions are evaluated using the objective function and stored in
        `self.solutions`.

        Returns
        -------
        None
            This method does not return any value. It initializes and stores solutions internally.

        Notes
        -----
        - The number of initial samples is determined by `self.solutions.fe_init`.
        - The sampled points are transformed from the normalized space to the original decision space.
        """
        self._solutions = Solution(self._config)
        x_init = lhs(self.dim, samples=self.fe_init)
        if self.no_partition:
            x_init = x_init * (self.ub - self.lb) + self.lb
        else:
            x_init = x_init * (self._partition[1] - self._partition[0]) + self._partition[0]

        y_init = np.array(self.evaluate(x_init))

        self.solutions._x_global = self.x_global
        if self.auto_update_solutions:
            # because the y_global is getting by self.evaluate(),
            # and the auto solution update also in self.evaluate(),
            # so we need disable auto_update_solutions temporarily
            # to set the y_global
            self.auto_update_solutions = False
            self.solutions._y_global = self.y_global
            self.auto_update_solutions = True
        else:
            self.solutions.append(x_init, y_init)
            self.solutions._y_global = self.y_global

    def random_uniform_x(self, size, within_partition=True) -> ndarray:
        """
        Sample points uniformly within the hypercube defined by :math:`(x_{lb}, x_{ub})^{dim}`.

        Parameters
        ----------
        size : int
            Number of samples, int
        within_partition : bool
            Within partition or not

        Returns
        -------
        samples : ndarray
            Samples
        """

        if not self.no_partition and within_partition:
            return np.random.uniform(self._partition[0], self._partition[1], size=(size, self.dim))
        else:
            return np.random.uniform(low=self.lb, high=self.ub, size=(size, self.dim))

    def normalize_x(self, points):
        """
        Normalize a set of points to the :math:`(0, 1)^D` space, where
        :math:`D` is the dimension of the problem's decision space.

        Parameters
        ----------
        points : ndarray or list or tuple
            A set of points.
        Returns
        -------
        points : ndarray
            Normalized points
        """
        points = FunctionInputs(x=points, dim=self.dim).x
        points = self.clip_x(points)
        return (points - self.lb) / (self.ub - self.lb)

    def denormalize_x(self, points):
        """
        Denormalize a set of points to function original decision space.
        You may want to set normalize=True if you wish the function's
        decision variables :math:`x \\in (0,1)^d`
        Parameters
        ----------
        points : ndarray or list or tuple
            A set of points. :math:`x \\in (0,1)^d`
        Returns
        -------
        points : ndarray
            Denormalized points
        """
        points = FunctionInputs(x=points, dim=self.dim).x
        points = np.clip(points, 0, 1)
        return points * (self.ub - self.lb) + self.lb

    def normalize_y(self, y):
        y = np.clip(y, self.solutions.y_min, self.solutions.y_max)
        return (y - self.solutions.y_min) / (self.solutions.y_max - self.solutions.y_min)

    def denormalize_y(self, y):
        y = np.clip(y, 0, 1)
        return y * (self.solutions.y_max - self.solutions.y_min) + self.solutions.y_min

    def transform_x(self, x):
        return self._transformer.transform_x(x)

    def clip_x(self, x):
        return np.clip(x, self.lb, self.ub)

    def inverse_transform_x(self, x):
        return self._transformer.inverse_transform_x(x)

    def evaluate(self, x: ndarray):
        _x = self.before_eval(x)
        y = np.apply_along_axis(self._eval_single, 1, _x)
        y = self.after_eval(x, y)
        return y

    def before_eval(self, x):
        _x = copy.deepcopy(x)
        _x = FunctionInputs(x=_x, dim=self.dim).x
        _x = self.transform_x(_x)
        _x = np.clip(_x, self.lb, self.ub)
        return _x.reshape(-1, self.dim)

    def after_eval(self, x, y):
        _x = x.reshape(-1, self.dim)
        _y = y.reshape(-1, self.obj)
        if self.auto_update_solutions:
            self.solutions.append(_x, _y)
        return _y

    @property
    def x_global(self):
        if self.is_known_optimal:
            return self.inverse_transform_x(self._x_global)
        else:
            return self._x_global

    @property
    def is_known_optimal(self):
        return self._x_global.size > 0

    @property
    def y_global(self):
        if self.is_known_optimal:
            return self.evaluate(self.x_global).squeeze()
        else:
            return np.array([])

    @property
    def shift(self):
        return self._transformer.shift

    @property
    def rotation(self):
        return self._transformer.rotation

    @abstractmethod
    def _eval_single(self, x: ndarray):
        """
        OjbFunc:= :math:`\\mathbf{x} \\to f \\to \\mathbf{y}, \\mathbf{x} \\in \\mathbb{R}^{D1},
        \\mathbf{y}\\in \\mathbb{R}^{D2}`, where :math:`D1` is the dimension of decision space
        and :math:`D2` is that of objective space.

        Parameters
        ----------
        x: ndarray
            Input vector of the problem.

        Returns
        -------
        y:
            Function value
        """

    @property
    def no_partition(self) -> bool:
        return not hasattr(self, "_partition")

    @property
    def id(self) -> int:
        return self._id if hasattr(self, "_id") else -1

    @property
    def lb(self) -> ndarray:
        return np.asarray(self._config.lb)

    @property
    def ub(self) -> ndarray:
        return np.asarray(self._config.ub)

    @property
    def name(self) -> str:
        return type(self).__name__

    @property
    def dim(self) -> int:
        return self._config.dim

    @property
    def obj(self) -> int:
        return self._config.obj

    @property
    def fe_init(self) -> int:
        return self._config.fe_init

    @property
    def fe_max(self) -> int:
        return self._config.fe_max

    @property
    def fe_available(self) -> int:
        return self.fe_max - self.solutions.size

    @property
    def npd(self) -> int:
        return self._config.npd

    @property
    def solutions(self) -> Solution:
        return self._solutions if hasattr(self, '_solutions') else Solution(self._config)


class MultiTaskProblem(ABC):
    """
    The following attributes can be config for a subclass:
        - is_realworld: bool
        - intro: str
        - notes: str
        - references: list[str]
    """
    is_realworld: bool
    intro = "Not set"
    notes = "Not set"
    references: list[str] = ["Not set"]

    def __init__(self, *args, **kwargs):
        """
        args and kwargs will be passed to the `_init_tasks` method.
        """
        self.seed = kwargs.pop('seed', 123)
        self.random_ctrl = kwargs.pop('random_ctrl', 'weak')
        _init_solutions = kwargs.pop('_init_solutions', True)
        self._problem = self.__init_tasks(*args, **kwargs)
        if _init_solutions:
            self._init_solutions()

    def __init_tasks(self, *args, **kwargs):
        problem = self._init_tasks(*args, **kwargs)
        if not isinstance(problem, (list, tuple)):
            raise TypeError(f"init_tasks() must return a list or tuple, got {type(problem)} instead.")
        self._check_problem(list(problem))
        return problem

    @staticmethod
    def _check_problem(problem: list[SingleTaskProblem]):
        if problem[0].id == -1:
            for i in range(len(problem)):
                problem[i].set_id(i + 1)

    @abstractmethod
    def _init_tasks(self, *args, **kwargs) -> Union[list[SingleTaskProblem], tuple[SingleTaskProblem, ...]]: ...

    def __iter__(self):
        return iter(self._problem)

    def __getitem__(self, index) -> Union[SingleTaskProblem, list[SingleTaskProblem]]:
        if isinstance(index, slice):
            return self._problem[index]
        elif isinstance(index, int):
            if index < 0:
                raise IndexError(f"Index {index} < 0 is invalid.")
            if index >= len(self._problem):
                raise IndexError(f"Index {index} > total tasks={len(self._problem)} in problem {self.name}")
            return self._problem[index]
        else:
            raise TypeError(f"Index must be an integer or a slice, but {type(index)} given.")

    def _init_solutions(self):
        if self.random_ctrl == 'no':
            self.__init_partition()
            self.__init_solutions()
        elif self.random_ctrl == 'weak':
            self._set_seed()
            self.__init_partition()
            self._unset_seed()
            self.__init_solutions()
        elif self.random_ctrl == 'strong':
            self._set_seed()
            self.__init_partition()
            self.__init_solutions()
            self._unset_seed()
        else:
            raise ValueError('Invalid random_ctrl option.')

    def __init_partition(self):
        for p in self._problem:
            p.init_partition()

    def __init_solutions(self):
        for p in self._problem:
            p.init_solutions()

    def _set_seed(self):
        self._rdm_state = np.random.get_state()
        np.random.seed(self.seed)

    def _unset_seed(self):
        np.random.set_state(self._rdm_state)

    def iplot_tasks_3d(
            self,
            tasks_id: Union[list, tuple, range] = (1, 2, 3, 4),
            shape: tuple[int, int] = (2, 2),
            dims=(0, 1),
            font_size: int = 10,
            cmap: Union[str, Cmaps] = Cmaps.viridis,
            color: Union[str, StrColors, tuple[float, float, float], None] = None,
            n_points: int = 100,
            scale_mode: Literal['xy', 'y'] = 'y',
            show_grid: bool = True,
    ):
        """
        Plot a set of tasks' 3d landscape in an interactive window.

        Parameters
        ----------
        tasks_id:
            The tasks id to be plotting.
        shape:
            The grid shape of multi-plots.
        cmap:
            Color map, use `pyfmto.utilities.Cmaps` to easily select alternative options.
        color:
            Use to make each mesh have a single solid color. ``cmap`` will be overridden
            if color is specified. Either a string, RGB list, or hex color string. For
            example: ``color='white'``, ``color=StrColors.white``, ``color='w'``,
            ``color=[1.0,1.0,1.0]``, or ``color='#FFFFFF'``.
        dims:
            The selected dims to be plotting.
        scale_mode:
            if pass 'xy', normalize both the decision space and objective space to (0, 1).
            if pass 'y', scale objective value to match the decision space, which will keep
            the original decision space information.
        n_points:
            Number of points in one dimension, total n_points**2 points.
        font_size:
            Fontsize of subplots title.
        show_grid:
            If show coordinates plane.
        """
        if np.any(np.array(tasks_id) <= 0) or np.any(np.array(tasks_id) > self.task_num):
            raise ValueError(f"tasks id should in range [1, {self.task_num}]")
        if len(tasks_id) > shape[0] * shape[1]:
            raise ValueError(f"The number of clients {len(tasks_id)} is larger than "
                             f"the number of cells in grid {shape[0]}*{shape[1]}")
        import pyvista as pv
        plotter = pv.Plotter(shape=shape)
        for index, tid in enumerate(tasks_id):
            row, col = index // shape[1], index % shape[1]
            plotter.subplot(row, col)
            task = self._problem[tid - 1]
            task.iplot_3d(
                dims=dims,
                n_points=n_points,
                scale_mode=scale_mode,
                cmap=cmap,
                color=color,
                show_grid=show_grid,
                font_size=font_size,
                plotter=plotter,
            )
        plotter.show()

    def plot_distribution(
            self,
            dims=(0, 1),
            style: Union[str, SeabornStyles] = SeabornStyles.whitegrid,
            figsize: tuple[float, float, float] = (11., 9., 1.),
            filename=None
    ):
        """
        Plot the distribution of initial solutions for each task in the multitask problem.

        This function visualizes how the initial solutions are distributed across the
        decision space for each task. It helps to understand the partitioning strategy
        and the diversity of initial samples among different tasks.

        Parameters
        ----------
        dims : tuple of int, default=(0, 1)
            The two dimensions to plot. Only 2D visualization is supported.

        style : str or SeabornStyles, default=SeabornStyles.whitegrid
            The seaborn style to use for the plot. Can be any valid seaborn style
            or a SeabornStyles enum value.

        figsize : tuple of float, default=(11., 9., 1.)
            Figure size specification as (width, height, scale_factor). The final
            figure size will be (width*scale_factor, height*scale_factor).

        filename : str, optional
            Path to save the figure. If None, the figure will be displayed directly.

        Notes
        -----
        - The function normalizes all solution points to [0, 1] space for consistent visualization.
        - Grid lines are shown according to the 'npd' setting to visualize partitions.
        - Each task is represented with a unique color and marker style.
        - The plot includes a legend indicating which color/marker corresponds to which task.

        Examples
        --------
        #>>> problem.plot_distribution()
        #>>> problem.plot_distribution(dims=(1, 2), filename='distribution.png')
        """
        init_x: dict[str, ndarray] = {f"T{p.id:02}({p.name})": p.normalize_x(p.solutions.x) for p in self}
        npd = self[0].npd
        tick_interval = 1.0 / npd
        ticks = np.arange(0, 1 + tick_interval, tick_interval)
        dim1 = min(dims)
        dim2 = max(dims)
        data = []
        for label, points in init_x.items():
            for point in points:
                data.append({
                    'x': point[dim1],
                    'y': point[dim2],
                    'problem': label
                })
        w, h, s = figsize
        plt.figure(figsize=(w * s, h * s))
        sns.set(style=str(style))
        sns.scatterplot(data=pd.DataFrame(data), x='x', y='y', hue='problem', style='problem', s=80)
        plt.title(f'Initial Solutions of Each Task (npd={npd})')
        plt.xlabel(f'X{dim1}')
        plt.ylabel(f'X{dim2}', rotation=0)
        plt.xticks(ticks)
        plt.yticks(ticks)
        plt.legend(title='Tasks', bbox_to_anchor=(1., 1), loc='upper left')
        plt.tight_layout()
        if filename:
            plt.savefig(filename)
        else:
            plt.show()

    def plot_similarity_heatmap(
            self,
            n_samples: int = 1000,
            method: Literal['kendalltau', 'spearmanr', 'pearsonr'] = 'spearmanr',
            p_value: float = 1.,
            figsize: tuple[float, float, float] = (9., 7., 1.),
            cmap: Optional[Union[str, Cmaps]] = None,
            masker: Optional[float] = np.nan,
            font_size: int = 10,
            fmt: str = '.2f',
            triu: Literal['full', 'lower', 'upper'] = 'full',
            vmax: Optional[float] = None,
            vmin: Optional[float] = None,
            linewidth: float = 0.5,
            filename: Optional[Union[str, Path]] = None,
    ):
        """
        Plot a heatmap showing the similarity between tasks.

        This function computes correlation coefficients between all pairs of tasks
        using sampled points in the decision space, and visualizes them as a heatmap.
        Correlations that are not statistically significant (based on p-value) can
        be masked out.

        Parameters
        ----------
        n_samples : int, default=1000
         Number of Latin Hypercube samples to use for evaluating the functions.

        method :
         Correlation method to use:
         'spearmanr': Spearman rank correlation;
         'kendalltau': Kendall tau correlation;
         'pearsonr': Pearson correlation coefficient

        p_value :
         Upper bound for p-value. Correlations with p-values greater than this
         value are considered not statistically significant and will be masked.

        figsize :
         Figure size as (width, height, scale).

        cmap :
         Colormap for the heatmap. If None, uses the default matplotlib colormap.

        masker :
         Value to use for masking non-significant correlations. Default is NaN.

        font_size :
         Font size for annotation text in heatmap cells.

        fmt :
         Format string for annotating values in heatmap cells.

        triu :
         Which part of the matrix to display:
         - 'full': Show full matrix
         - 'lower': Show only lower triangular matrix
         - 'upper': Show only upper triangular matrix

        vmax :
         Maximum value for colormap normalization.

        vmin :
         Minimum value for colormap normalization.

        linewidth :
         Width of lines separating heatmap cells.

        filename :
         File path to save the plot. If None, displays the plot instead.

        Notes
        -----
        The heatmap shows correlation coefficients between tasks. Non-significant
        correlations (based on the p_ub threshold) are masked with the masker value.
        The diagonal always shows 1.0 as each task is perfectly correlated with itself.
        Task labels are formatted as 'T{ID:02}' where ID is the task identifier.

        See Also
        --------
        See `scipy.stats.kendalltau`, `spearmanr`, or `pearsonr` documentation for more detail.
        """
        methods = {
            'spearmanr': spearmanr,
            'kendalltau': kendalltau,
            'pearsonr': pearsonr,
        }
        if method in methods:
            corr = methods[method]
        else:
            raise ValueError(f"Invalid method: {method}, support spearmanr, kendalltau, and pearsonr")
        _cmap = None if cmap is None else str(cmap)
        x = lhs(self[0].dim, samples=n_samples)
        evals = [f.evaluate(f.denormalize_x(x)).squeeze() for f in self]
        evals = np.array(evals)
        _sig = [[[*corr(evals[i], evals[j])] for j in range(self.task_num)] for i in range(self.task_num)]
        significance = np.array(_sig)
        statis = significance[:, :, 0]
        pvalue = significance[:, :, 1]
        triu_mask = np.triu(np.ones_like(statis, dtype=bool), k=1)
        statis[pvalue > p_value] = masker if masker else np.nan
        if triu == 'lower':
            statis[~triu_mask] = np.nan
        elif triu == 'upper':
            statis[triu_mask] = np.nan
        cols = [f"T{f.id:02}" for f in self]
        df = pd.DataFrame(statis[::-1], columns=cols, index=cols[::-1])
        w, h, s = figsize
        _, ax = plt.subplots(figsize=(w * s, h * s))
        sns.heatmap(
            data=df,
            cmap=_cmap,
            annot=True,
            fmt=fmt,
            linewidths=linewidth,
            ax=ax,
            annot_kws={"size": font_size},
            vmax=vmax,
            vmin=vmin,
        )
        ax.set_yticklabels(ax.get_yticklabels(), rotation=0)
        plt.tight_layout()
        if filename:
            plt.savefig(filename)
        else:
            plt.show()
        plt.close()

    @property
    def name(self) -> str:
        return self.__class__.__name__

    @property
    def task_num(self):
        return len(self._problem)

    def get_info(self) -> dict:
        return {
            "ProbName": [self.name],
            "ProbType": ['Realworld' if self.is_realworld else 'Synthetic'],
            "TaskNum": [len(self._problem)],
            "DecDim": [self._problem[0].dim],
            "ObjDim": [self._problem[0].obj]
        }

    def __len__(self):
        return len(self._problem) if self._problem is not None else 0

    def __repr__(self):
        t = "realworld" if self.is_realworld else "synthetic"
        return f"{self.name}({len(self._problem)} {t} tasks)"

    def __str__(self):
        def _reformat_lines(text, width=80):
            _dedent = textwrap.dedent(text)
            _width = textwrap.fill(_dedent, width=width)
            _indent = textwrap.indent(_width, ' ' * 4)
            if not _indent.startswith('\n'):
                _indent = '\n' + _indent
            if not _indent.endswith('\n'):
                _indent += '\n'
            return _indent

        tab = tabulate(self.get_info(), headers="keys", tablefmt="rounded_grid")
        tab = textwrap.indent(tab, " " * 8)
        tab_width = tab.find('\n') + 8
        prob_name = self.name.center(tab_width - 16, '=')
        prob_name = textwrap.indent(prob_name, " " * 8)
        intro = _reformat_lines(self.intro, tab_width)
        notes = _reformat_lines(self.notes, tab_width)
        ref = [_reformat_lines(txt, tab_width) for txt in self.references]
        ref_str = "\n".join(ref)

        return (
            f"{prob_name}\n"
            f"{tab}\n"
            f"Introduction:{intro}\n"
            f"Notes:{notes}\n"
            f"References:{ref_str}"
        ).replace('\n\n\n', '\n\n')



from .benchmarks import (
    Ackley,
    Sphere,
    Griewank,
    Schwefel,
    Ellipsoid,
    Rastrigin,
    Rosenbrock,
    Weierstrass,
)


__all__ = [
    'Ackley',
    'Sphere',
    'Griewank',
    'Schwefel',
    'Ellipsoid',
    'Rastrigin',
    'Rosenbrock',
    'Weierstrass',
]



import unittest
import os
import sys
from pathlib import Path
from unittest.mock import Mock
from unittest.mock import patch

from pyfmto import list_problems, list_algorithms
from pyfmto.experiments import list_report_formats
from pyfmto.utilities.cli import update_path, main
from tests.helpers import gen_algorithm


class TestUpdatePath(unittest.TestCase):
    def setUp(self):
        self.original_sys_path = sys.path.copy()
        self.original_cwd = os.getcwd()
        self.temp_dir = Path().cwd() / "tmp"
        self.temp_dir.mkdir(parents=True, exist_ok=True)

    def tearDown(self):
        sys.path[:] = self.original_sys_path
        os.chdir(self.original_cwd)
        if self.temp_dir.exists():
            import shutil
            shutil.rmtree(self.temp_dir)

    def test_update_path_already_in_path(self):
        algorithms_dir = self.temp_dir / "algorithms"
        algorithms_dir.mkdir()

        with patch('pyfmto.utilities.cli.Path.cwd', return_value=self.temp_dir):
            if str(self.temp_dir) not in sys.path:
                sys.path.append(str(self.temp_dir))
            orig_len = len(sys.path)
            update_path()
            self.assertEqual(len(sys.path), orig_len)

    def test_update_path_path_normalization(self):
        algorithms_dir = self.temp_dir / "algorithms"
        algorithms_dir.mkdir()

        with patch('pyfmto.utilities.cli.Path.cwd', return_value=self.temp_dir):
            if str(self.temp_dir) in sys.path:
                sys.path.remove(str(self.temp_dir))
            update_path()
            self.assertIn(str(self.temp_dir), sys.path)
            path_index = sys.path.index(str(self.temp_dir))
            self.assertIsInstance(sys.path[path_index], str)


class TestMainFunction(unittest.TestCase):

    def setUp(self):
        self.algorithms_dir = Path().cwd() / "algorithms"
        self.algorithms_dir.mkdir(parents=True, exist_ok=True)
        self.original_argv = sys.argv

    def tearDown(self):
        sys.argv = self.original_argv
        if self.algorithms_dir.exists():
            import shutil
            shutil.rmtree(self.algorithms_dir)

    @patch('pyfmto.utilities.cli.Launcher')
    @patch('pyfmto.utilities.cli.Reports')
    def test_run_command(self, mock_reports, mock_launcher):
        # Setup mock launcher
        mock_launcher_instance = Mock()
        mock_launcher.return_value = mock_launcher_instance

        # Set up command line arguments
        test_args = ['pyfmto', 'run', '--config', 'test_config.yaml']
        with patch.object(sys, 'argv', test_args):
            main()

        # Verify launcher was created with correct config and run was called
        mock_launcher.assert_called_once_with(conf_file='test_config.yaml')
        mock_launcher_instance.run.assert_called_once()
        mock_reports.assert_not_called()

    @patch('pyfmto.utilities.cli.Launcher')
    @patch('pyfmto.utilities.cli.Reports')
    def test_report_command(self, mock_reports, mock_launcher):
        # Setup mock reports
        mock_reports_instance = Mock()
        mock_reports.return_value = mock_reports_instance

        # Set up command line arguments
        test_args = ['pyfmto', 'report', '--config', 'test_config.yaml']
        with patch.object(sys, 'argv', test_args):
            main()

        # Verify reports was created with correct config and generate was called
        mock_reports.assert_called_once_with(conf_file='test_config.yaml')
        mock_reports_instance.generate.assert_called_once()
        mock_launcher.assert_not_called()

    @patch('pyfmto.utilities.cli.Launcher')
    @patch('pyfmto.utilities.cli.Reports')
    def test_default_config_file(self, mock_reports, mock_launcher):
        # Setup mock launcher
        mock_launcher_instance = Mock()
        mock_launcher.return_value = mock_launcher_instance

        # Set up command line arguments without config
        test_args = ['pyfmto', 'run']
        with patch.object(sys, 'argv', test_args):
            main()

        # Verify launcher was created with default config
        mock_launcher.assert_called_once_with(conf_file='config.yaml')
        mock_launcher_instance.run.assert_called_once()

    def test_list_command(self):
        options = ['algorithms', 'problems', 'reports']
        args_lst = [['pyfmto', 'list', option] for option in options]
        for test_args in args_lst:
            with self.subTest(test_args=test_args):
                with patch.object(sys, 'argv', test_args):
                    main()

    def test_show_command(self):
        gen_algorithm('Algor')
        gen_algorithm('ALG')
        options = {
            'prob': list(list_problems().keys()),
            'report': list_report_formats(),
            'alg': list(list_algorithms().keys()),
            'invalid': ['Invalid']
        }
        for grp, lst in options.items():
            for option in lst:
                with self.subTest(grp=grp, option=option):
                    test_args = ['pyfmto', 'show', f"{grp}.{option}"]
                    with patch.object(sys, 'argv', test_args):
                        main()



# Reference:
#   https://www.bilibili.com/video/BV1sK4y1x7e1
#   https://www.cnblogs.com/kangshuaibo/p/14700833.html

import logging.config
import logging.handlers
import time
from pathlib import Path
from logging.handlers import RotatingFileHandler

__all__ = ['logger']
Path('out', 'logs').mkdir(parents=True, exist_ok=True)


class PyfmtoRotatingFileHandler(RotatingFileHandler):
    def rotation_filename(self, default_name: str) -> str:
        filename = Path(self.baseFilename)
        base = filename.stem
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
        new_name = filename.with_name(f"{base} {timestamp}.log")
        return str(new_name)


LOG_CONF = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'simpleFormatter': {
            'format': '%(levelname)-8s%(asctime)-22s%(filename)16s->line(%(lineno)s)|%(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        }
    },
    'handlers': {
        'pyfmto_handler': {
            '()': PyfmtoRotatingFileHandler,
            'level': 'DEBUG',
            'formatter': 'simpleFormatter',
            'filename': 'out/logs/pyfmto.log',
            'maxBytes': 2 * 1024 * 1024,
            'backupCount': 10
        }
    },
    'loggers': {
        'pyfmto': {
            'level': 'DEBUG',
            'handlers': ['pyfmto_handler'],
            'propagate': 0
        }
    }
}


logging.config.dictConfig(LOG_CONF)
logger = logging.getLogger('pyfmto')



from .client import Client, record_runtime
from .server import Server
from .packages import ClientPackage, SyncDataManager

__all__ = [
    'Client',
    'Server',
    'SyncDataManager',
    'ClientPackage',
    'record_runtime'
]



import msgpack
import numpy as np
from pathlib import Path
from typing import Union, Any, Optional

from pydantic import validate_call
from ruamel.yaml import YAML, CommentedMap
from ruamel.yaml.error import MarkedYAMLError

yaml = YAML()
yaml.default_flow_style = False
yaml.preserve_quotes = True
yaml.indent(mapping=2, sequence=4, offset=2)

T_Path = Union[str, Path]

__all__ = [
    'load_yaml',
    'save_yaml',
    'parse_yaml',
    'dumps_yaml',
    'load_msgpack',
    'save_msgpack',
]


@validate_call
def load_yaml(filename: T_Path):
    with open(filename, 'r') as f:
        return parse_yaml(f.read())


def save_yaml(data: Union[dict, CommentedMap], filename: T_Path):
    with open(filename, 'w') as f:
        f.write(dumps_yaml(data))


@validate_call
def parse_yaml(text: Optional[str]):
    if not text:
        return {}
    lines = [line for line in text.splitlines() if line.strip() != '']
    try:
        return yaml.load('\n'.join(lines))
    except MarkedYAMLError:
        raise


def dumps_yaml(data: Union[dict, CommentedMap]):
    from io import StringIO
    string_stream = StringIO()
    yaml.dump(data, string_stream)
    text: list[str] = []
    rows = string_stream.getvalue().splitlines()

    add_newline = False
    for line in rows:
        if line.startswith(' '):
            add_newline = True
            break
    if not add_newline:
        return '\n'.join(rows)

    for line in rows:
        if text and not line.startswith(' '):
            # Only add a newline if the line is
            # not indented and not the first line
            text.append(f"\n{line}")
        else:
            text.append(line)
    return '\n'.join(text)


@validate_call
def save_msgpack(data: dict, filename: T_Path) -> None:
    with open(filename, 'wb') as f:
        packed = msgpack.packb(data, default=_encode_hook, use_bin_type=True)
        f.write(packed)


@validate_call
def load_msgpack(filename: T_Path) -> dict:
    with open(filename, 'rb') as f:
        data = msgpack.unpackb(f.read(), object_hook=_decode_hook, raw=False, strict_map_key=False)
    return data


def _encode_hook(obj: Any) -> dict:
    if isinstance(obj, np.ndarray):
        return {
            "__ndarray__": True,
            "dtype": obj.dtype.name,
            "shape": obj.shape,
            "data": obj.tobytes()
        }
    elif isinstance(obj, set):
        return {
            "__set__": True,
            "items": list(obj)
        }
    else:
        raise TypeError(f"Unsupported type: {type(obj)}")


def _decode_hook(obj: dict) -> Any:
    if "__ndarray__" in obj:
        dtype = np.dtype(obj["dtype"])
        shape = tuple(obj["shape"])
        data = obj["data"]
        return np.frombuffer(data, dtype=dtype).reshape(shape)
    elif "__set__" in obj:
        return set(obj["items"])
    return obj



from enum import Enum

__all__ = [
    'Cmaps',
    'StrColors',
    'StrOptions',
    'SeabornStyles',
    'SeabornPalettes',
]


class StrOptions(str, Enum):
    def __str__(self):
        return self.name


class Cmaps(StrOptions):
    """
    Matplotlib colormap options.

    This class provides a set of predefined colormap names with their
    descriptions for use in matplotlib plots.

    Uniform
    -------
    - ``viridis`` : Purple -> Yellow (colorblind-friendly)"
    - ``plasma``  : Purple -> Pink -> Yellow"
    - ``inferno`` : Black -> Red -> Yellow"
    - ``magma``   : Black -> Purple -> White"
    - ``cividis`` : Colorblind-friendly version of viridis"

    Sequential
    ----------
    - ``Greys``   : White to black grayscale"
    - ``Purples`` : White to purple sequential colormap"
    - ``Blues``   : White to blue sequential colormap"
    - ``Greens``  : White to green sequential colormap"
    - ``Oranges`` : White to orange sequential colormap"
    - ``Reds``    : White to red sequential colormap"
    - ``YlOrBr``  : Yellow -> Orange -> Brown sequential"
    - ``YlOrRd``  : Yellow -> Orange -> Red sequential"
    - ``OrRd``    : Orange -> Red sequential"
    - ``PuRd``    : Purple -> Red sequential"
    - ``RdPu``    : Red -> Purple sequential"
    - ``BuPu``    : Blue -> Purple sequential"
    - ``GnBu``    : Green -> Blue sequential"
    - ``PuBu``    : Purple -> Blue sequential"
    - ``YlGnBu``  : Yellow -> Green -> Blue sequential"
    - ``PuBuGn``  : Purple -> Blue -> Green sequential"
    - ``BuGn``    : Blue -> Green sequential"
    - ``YlGn``    : Yellow -> Green sequential"

    Diverging
    ---------
    - ``PiYG``     : Pink -> Yellow -> Green"
    - ``PRGn``     : Purple -> Rose -> Green"
    - ``BrBG``     : Brown -> Beige -> Green"
    - ``PuOr``     : Orange -> White -> Purple"
    - ``RdGy``     : Red -> Grey -> Blue"
    - ``RdBu``     : Red -> White -> Blue"
    - ``RdYlBu``   : Red -> Yellow -> Blue"
    - ``RdYlGn``   : Red -> Yellow -> Green"
    - ``Spectral`` : Red -> Orange -> Blue -> Green"
    - ``coolwarm`` : Blue -> Grey -> Red"
    - ``bwr``      : Blue -> White -> Red"
    - ``seismic``  : Blue -> White -> Red"

    Qualitative
    -----------
    - ``Pastel1`` : Pastel qualitative palette"
    - ``Pastel2`` : Second pastel qualitative palette"
    - ``Paired``  : Paired qualitative palette"
    - ``Accent``  : Accent qualitative palette"
    - ``Dark2``   : Dark qualitative palette"
    - ``Set1``    : Distinct qualitative palette"
    - ``Set2``    : Second distinct qualitative palette"
    - ``Set3``    : Third distinct qualitative palette"
    - ``tab10``   : Tableau 10-color palette"
    - ``tab20``   : Tableau 20-color palette"
    - ``tab20b``  : Tableau 20-color bright palette"
    - ``tab20c``  : Tableau 20-color medium palette"

    Miscellaneous
    -------------
    - ``terrain``   : Terrain-like colormap"
    - ``ocean``     : Ocean-like colormap"
    - ``cubehelix`` : Cube helix colormap"
    - ``rainbow``   : Rainbow colormap"
    - ``twilight``  : Twilight colormap"

    Examples
    --------
        >>> from pyfmto.utilities.stroptions import Cmaps
        >>> print(Cmaps.viridis)
        viridis
        >>> plot_func_2d(func, lb, ub, dim, cmap=Cmaps.viridis)
    """

    viridis = "[Uniform] Purple -> Yellow (colorblind-friendly)"
    plasma = "[Uniform] Purple -> Pink -> Yellow"
    inferno = "[Uniform] Black -> Red -> Yellow"
    magma = "[Uniform] Black -> Purple -> White"
    cividis = "[Uniform] Colorblind-friendly version of viridis"

    Greys = "[Sequential] White to black grayscale"
    Purples = "[Sequential] White to purple sequential colormap"
    Blues = "[Sequential] White to blue sequential colormap"
    Greens = "[Sequential] White to green sequential colormap"
    Oranges = "[Sequential] White to orange sequential colormap"
    Reds = "[Sequential] White to red sequential colormap"
    YlOrBr = "[Sequential] Yellow -> Orange -> Brown sequential"
    YlOrRd = "[Sequential] Yellow -> Orange -> Red sequential"
    OrRd = "[Sequential] Orange -> Red sequential"
    PuRd = "[Sequential] Purple -> Red sequential"
    RdPu = "[Sequential] Red -> Purple sequential"
    BuPu = "[Sequential] Blue -> Purple sequential"
    GnBu = "[Sequential] Green -> Blue sequential"
    PuBu = "[Sequential] Purple -> Blue sequential"
    YlGnBu = "[Sequential] Yellow -> Green -> Blue sequential"
    PuBuGn = "[Sequential] Purple -> Blue -> Green sequential"
    BuGn = "[Sequential] Blue -> Green sequential"
    YlGn = "[Sequential] Yellow -> Green sequential"

    PiYG = "[Diverging] Pink -> Yellow -> Green"
    PRGn = "[Diverging] Purple -> Rose -> Green"
    BrBG = "[Diverging] Brown -> Beige -> Green"
    PuOr = "[Diverging] Orange -> White -> Purple"
    RdGy = "[Diverging] Red -> Grey -> Blue"
    RdBu = "[Diverging] Red -> White -> Blue"
    RdYlBu = "[Diverging] Red -> Yellow -> Blue"
    RdYlGn = "[Diverging] Red -> Yellow -> Green"
    Spectral = "[Diverging] Red -> Orange -> Blue -> Green"
    coolwarm = "[Diverging] Blue -> Grey -> Red"
    bwr = "[Diverging] Blue -> White -> Red"
    seismic = "[Diverging] Blue -> White -> Red"

    Pastel1 = "[Qualitative] Pastel qualitative palette"
    Pastel2 = "[Qualitative] Second pastel qualitative palette"
    Paired = "[Qualitative] Paired qualitative palette"
    Accent = "[Qualitative] Accent qualitative palette"
    Dark2 = "[Qualitative] Dark qualitative palette"
    Set1 = "[Qualitative] Distinct qualitative palette"
    Set2 = "[Qualitative] Second distinct qualitative palette"
    Set3 = "[Qualitative] Third distinct qualitative palette"
    tab10 = "[Qualitative] Tableau 10-color palette"
    tab20 = "[Qualitative] Tableau 20-color palette"
    tab20b = "[Qualitative] Tableau 20-color bright palette"
    tab20c = "[Qualitative] Tableau 20-color medium palette"

    terrain = "[Miscellaneous] Terrain-like colormap"
    ocean = "[Miscellaneous] Ocean-like colormap"
    cubehelix = "[Miscellaneous] Cube helix colormap"
    rainbow = "[Miscellaneous] Rainbow colormap"
    twilight = "[Miscellaneous] Twilight colormap"


class SeabornPalettes(StrOptions):
    """
    Seaborn palette options.

    This class provides a set of predefined palette names with their
    descriptions for use in seaborn plots.

    Available palettes
    ------------------
    - ``deep`` : Deep color palette with 10 colors
    - ``muted`` : Muted color palette with 10 colors
    - ``bright`` : Bright color palette with 10 colors
    - ``pastel`` : Pastel color palette with 10 colors
    - ``dark`` : Dark color palette with 10 colors
    - ``colorblind`` : Colorblind-friendly palette with 10 colors
    - ``husl`` : HUSL color space palette
    - ``Set1`` : Brewer Set1 palette
    - ``Set2`` : Brewer Set2 palette
    - ``Set3`` : Brewer Set3 palette
    - ``Paired`` : Brewer Paired palette
    - ``viridis`` : Viridis sequential colormap
    - ``plasma`` : Plasma sequential colormap
    - ``inferno`` : Inferno sequential colormap
    - ``magma`` : Magma sequential colormap

    Examples
    --------
        >>> from pyfmto.utilities.stroptions import SeabornPalettes
        >>> print(SeabornPalettes.deep)
        deep
        >>> sns.scatterplot(data=df, x='x', y='y', hue='category', palette=SeabornPalettes.deep)
    """

    # Seaborn default palettes
    deep = "Deep color palette with 10 colors"
    muted = "Muted color palette with 10 colors"
    bright = "Bright color palette with 10 colors"
    pastel = "Pastel color palette with 10 colors"
    dark = "Dark color palette with 10 colors"
    colorblind = "Colorblind-friendly palette with 10 colors"

    # Other categorical palettes
    husl = "HUSL color space palette"
    Set1 = "Brewer Set1 palette"
    Set2 = "Brewer Set2 palette"
    Set3 = "Brewer Set3 palette"
    Paired = "Brewer Paired palette"

    # Sequential palettes (from matplotlib colormaps)
    viridis = "Viridis sequential colormap"
    plasma = "Plasma sequential colormap"
    inferno = "Inferno sequential colormap"
    magma = "Magma sequential colormap"


class SeabornStyles(StrOptions):
    """
    Seaborn style options.

    This class provides a set of predefined style names with their
    descriptions for use in seaborn plots.

    Available styles
    -----
    - ``darkgrid`` : Dark background with grid lines (default seaborn style)
    - ``whitegrid`` : White background with grid lines
    - ``dark`` : Dark background with no grid lines
    - ``white`` : White background with no grid lines
    - ``ticks`` : White background with ticks on the axes

    Examples
    --------
        >>> from pyfmto.utilities.stroptions import SeabornStyles
        >>> print(SeabornStyles.whitegrid)
        whitegrid
        >>> sns.set_style(str(SeabornStyles.whitegrid))
    """

    darkgrid = "Dark background with grid lines (default seaborn style)"
    whitegrid = "White background with grid lines"
    dark = "Dark background with no grid lines"
    white = "White background with no grid lines"
    ticks = "White background with ticks on the axes"


class StrColors(StrOptions):
    black = 'black'
    gray = 'gray'
    red = 'red'
    green = 'green'
    yellow = 'yellow'
    blue = 'blue'
    magenta = 'magenta'
    cyan = 'cyan'
    white = 'white'



import numpy as np
from numpy import ndarray

from ..problem import MultiTaskProblem as Mtp
from ..benchmarks import (
    Ackley,
    Griewank,
    Rastrigin,
    Rosenbrock,
    Schwefel,
    Sphere,
    Weierstrass)

__all__ = ['Tetci2019']


class Tetci2019(Mtp):
    """
    dim: 10  # 1 <= dim <= 50
    """
    is_realworld = False
    intro = ""
    notes = """
        The number of tasks is either 10 or 8, depending on the input dimension.
        If dim <= 25, 10 tasks.
        If dim > 25, 8 tasks (due to Weierstrass function only support 1 <= dim <= 25)
        All tasks share the same dimensionality.
        Tasks differ by shift vectors applied to their base functions.
    """
    references = [
        """
        CHEN Y, ZHONG J, FENG L, et al. An Adaptive Archive-Based Evolutionary
        Framework for Many-Task Optimization[J/OL]. IEEE TETCI, 2020: 369-384.
        DOI:10.1109/tetci.2019.2916051.
        """
    ]

    def __init__(self, dim=10, **kwargs):
        if not 1 <= dim <= 50:
            raise ValueError('dim must be in [1, 50]')
        super().__init__(dim, **kwargs)

    def get_info(self):
        category = ['Easy'] * 4 + ['Complex'] * 6
        assisted_task = [None] * 4 + ['T1', 'T2', 'T3,T4', None, 'T4', None]
        if self.task_num == 8:
            del category[6], category[3]
            del assisted_task[6], assisted_task[3]
        return {
            "TaskID": [t.id for t in self._problem],
            "TaskName": [t.name for t in self._problem],
            "DecDim": [t.dim for t in self._problem],
            "Lower": [t.lb[0] for t in self._problem],
            "Upper": [t.ub[0] for t in self._problem],
            "Category": category,
            "Assisted By": assisted_task
        }

    def _init_tasks(self, dim, **kwargs):
        f1 = Sphere(dim, lb=-100, ub=100, **kwargs)
        f2 = Sphere(dim, lb=-100, ub=100, **kwargs)
        f3 = Sphere(dim, lb=-100, ub=100, **kwargs)
        f4 = Weierstrass(dim, lb=-0.5, ub=0.5, **kwargs)
        f5 = Rosenbrock(dim, lb=-50, ub=50, **kwargs)  # Ideal Assisted Task f1
        f6 = Ackley(dim, lb=-50, ub=50, **kwargs)  # Ideal Assisted Task f2
        f7 = Weierstrass(dim, lb=-0.4, ub=0.4, **kwargs)  # Ideal Assisted Task f3,f4
        f8 = Schwefel(dim, lb=-500, ub=500, **kwargs)
        f9 = Griewank(dim, lb=-100, ub=100, **kwargs)  # Ideal Assisted Task f4
        f10 = Rastrigin(dim, lb=-50, ub=50, **kwargs)

        functions = [f1, f2, f3, f4, f5, f6, f7, f8, f9, f10]
        [functions[idx].set_id(idx + 1) for idx in range(len(functions))]

        s9 = np.ones(dim)
        s9[:dim // 2] = -80
        s9[s9 == 1] = 80

        s10 = np.ones(dim)
        s10[: dim // 2] = 40
        s10[s10 == 1] = -40

        # Here, we set the f8 shift value to 0; if we set it
        # to 420.9687, which is the setting from the reference
        # paper and also represents the global optimality of
        # f8, the global optimum will fall outside the search
        # space (-500, 500).
        shifts = [0, 80, -80, -0.4, 0, 40, -0.4, 0, s9, s10]

        if dim > 25:
            del functions[6], functions[3]
            del shifts[6], shifts[3]

        for func, shift in zip(functions, shifts):
            if not isinstance(shift, ndarray):
                func.set_transform(shift=np.ones(dim) * shift)
            else:
                func.set_transform(shift=shift)

        return functions



import os
import platform
import subprocess

from tabulate import tabulate
from typing import Literal

__all__ = [
    'colored',
    'clear_console',
    'titled_tabulate',
    'tabulate_formats',
    'matched_str_head',
    'terminate_popen',
]


class TabulatesFormats:
    plain = 'plain'
    simple = 'simple'
    grid = 'grid'
    simple_grid = 'simple_grid'
    rounded_grid = 'rounded_grid'
    heavy_grid = 'heavy_grid'
    mixed_grid = 'mixed_grid'
    double_grid = 'double_grid'
    fancy_grid = 'fancy_grid'
    outline = 'outline'
    simple_outline = 'simple_outline'
    rounded_outline = 'rounded_outline'
    mixed_outline = 'mixed_outline'
    double_outline = 'double_outline'
    fancy_outline = 'fancy_outline'
    pipe = 'pipe'
    presto = 'presto'
    orgtbl = 'orgtbl'
    rst = 'rst'
    mediawiki = 'mediawiki'
    html = 'html'
    latex = 'latex'
    latex_raw = 'latex_raw'
    latex_booktabs = 'latex_booktabs'
    latex_longtable = 'latex_longtable'


tabulate_formats = TabulatesFormats()


def terminate_popen(process: subprocess.Popen):
    if process.stdout:
        process.stdout.close()
    if process.stderr:
        process.stderr.close()

    process.terminate()
    try:
        process.wait(timeout=5)
    except subprocess.TimeoutExpired:
        process.kill()
        process.wait()


def colored(text: str, color: Literal['black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan', 'white', 'reset']):
    color_map = {
        'black': '\033[30m',
        'red': '\033[31m',
        'green': '\033[32m',
        'yellow': '\033[33m',
        'blue': '\033[34m',
        'magenta': '\033[35m',
        'cyan': '\033[36m',
        'white': '\033[37m',
        'reset': '\033[0m'
    }

    if color not in color_map:
        raise ValueError(f"Unsupported color: {color}")

    return f"{color_map[color]}{text}{color_map['reset']}"


def titled_tabulate(title: str, fill_char: str, *args, **kwargs):
    title = ' ' + title if not title.startswith(' ') else title
    title = title + ' ' if not title.endswith(' ') else title
    tab = tabulate(*args, **kwargs)
    tit = title.center(tab.find('\n'), fill_char)
    return f"\n{tit}\n{tab}"


def clear_console():
    if platform.system() == 'Windows':
        os.system('cls')
    else:
        os.system('clear')


def matched_str_head(s: str, str_list: list[str]) -> str:
    for item in str_list:
        if item.startswith(s):
            return item
    return ''



from .reporter import Reports
from .launcher import Launcher
from .utils import (
    RunSolutions,
)

from ..utilities import parse_yaml, dumps_yaml, matched_str_head

__all__ = [
    'Reports',
    'Launcher',
    'RunSolutions',
    'show_default_conf',
    'list_report_formats',
]


def list_report_formats(print_it=False):
    res = [r[3:] for r in dir(Reports) if 'to_' in r]
    if print_it:
        print('\n'.join(res))
    return res


def show_default_conf(name: str):
    conf = parse_yaml(DEFAULT_CONF)
    full_name = matched_str_head(name, list(conf.keys()))
    if full_name:
        print(dumps_yaml(conf[full_name]))
    else:
        print(f"No matched format name {name} in report formats {list(conf.keys())}")


DEFAULT_CONF = """
curve:
    suffix: .png  # File extension for the generated curve image. options: '.png', '.jpg', '.eps', '.svg', '.pdf'
    quality: 3  # int in [1, 9]. Image quality parameter, affecting the quality of scalar images.
    showing_size: -1  # Use 'last showing_size data points' to plot curve. If -1, use all data points.
    merge: True  # Merge all plotted images to a single figure or not. Only support for '.png' and '.jpg' files.
    clear: True  # If True, clear plots after merge. Only applicable when merge is True.
    on_log_scale: False  # The plot is generated on a logarithmic scale.
    alpha: 0.2  # Transparency of the Standard Error region, 0 to 1 (completely to opaque).
violin:
    suffix: .png  # File extension for the generated violin image. Options: '.png', '.jpg', '.eps','.svg', '.pdf'.
    quality: 3   # int in [1, 9]. Image quality parameter, affecting the quality of scalar images.
    merge: bool  # Merge all plotted images to a single figure or not. Only support for '.png' and '.jpg' files.
    clear: True  # If True, clear plots after merge. Only applicable when merge is True.
console:
    pvalue: 0.5 # T-test threshold for determining statistical significance.
excel:
    pvalue: 0.5 # T-test threshold for determining statistical significance.
latex:
    pvalue: 0.5 # T-test threshold for determining statistical significance.
"""



import numpy as np
from pathlib import Path
from scipy.io import loadmat
from sklearn import svm
from sklearn.metrics import roc_auc_score
from sklearn.model_selection import train_test_split

from ..problem import (MultiTaskProblem as Mtp,
                       SingleTaskProblem as Stp)

__all__ = ["SvmLandmine"]


class _SingleSvmProblem(Stp):
    def __init__(self, dataset: tuple, **kwargs):
        super().__init__(dim=2, obj=1, lb=np.array([1e-4, 1e-2]), ub=np.array([10.0, 10.0]), **kwargs)
        self.x_train, self.x_test, self.y_train, self.y_test = dataset
        self.data_size = len(self.x_train)
        self.set_x_global(None)

    def evaluate(self, x):
        # training svm and get the prediction with real datasets
        _x = self.before_eval(x)
        y = np.array([self._eval_single(xi) for xi in _x])
        y = self.after_eval(x, y)
        return y

    def _eval_single(self, x):
        clf = svm.SVC(kernel="rbf", C=x[0], gamma=x[1], probability=True)
        clf.fit(self.x_train, self.y_train)
        pred = clf.predict_proba(self.x_test)
        score = roc_auc_score(self.y_test, pred[:, 1])
        return 1 - score


class SvmLandmine(Mtp):
    is_realworld = True
    intro = """
        A multitask Svm Landmine Detection hyperparameter optimization problem based on the Landmine dataset.
        Each task corresponds to training an SVM model with RBF kernel, optimizing hyperparameters
        (C and gamma) to minimize the classification error (1 - AUC score).
        """
    references = [
        """
        Qi, Y., Liu, D., Dunson, D., & Carin, L. (2008). Multitask compressive sensing
        with Dirichlet process priors. Proceedings of the 25th International Conference
        on Machine Learning - ICML'08, 768–775. https://doi.org/10.1145/1390156.1390253.
        """,
        """
        Zhu, H., Wang, X., & Jin, Y. (2023). Federated Many-Task Bayesian
        Optimization. IEEE Transactions on Evolutionary Computation, 1–1.
        https://doi.org/10.1109/TEVC.2023.3279775
        """
    ]

    def __init__(self, **kwargs):
        super().__init__(True, **kwargs)

    def _init_tasks(self, *args, **kwargs):
        dataset_path = Path(__file__).parents[1] / 'datasets' / 'svm_landmine' / "LandmineData.mat"
        landmine_data = loadmat(str(dataset_path))
        features = landmine_data["feature"][0]
        label = landmine_data["label"][0]

        tasks = []
        for i in range(len(features)):
            dataset = train_test_split(features[i], label[i], test_size=0.5, stratify=label[i], random_state=0)
            dataset = (dataset[0], dataset[1], dataset[2].squeeze(), dataset[3].squeeze())
            t = _SingleSvmProblem(dataset, **kwargs)
            t.set_id(i + 1)
            tasks.append(t)
        return tasks

    def get_info(self):
        return {
            "DecDim": [2],
            "Params": ["C, gamma"],
            "Lower": ['[1e-4, 1e-2]'],
            "Upper": ['[10, 10]']
        }



import asyncio
import logging
import pickle
import time
import traceback
import uvicorn
from abc import abstractmethod, ABC
from collections import defaultdict
from fastapi import FastAPI, Response, Depends, Request
from setproctitle import setproctitle
from tabulate import tabulate
from typing import final, Any

from .packages import ClientPackage, Actions
from pyfmto.utilities import logger

app = FastAPI()


async def load_body(request: Request):
    raw_data = await request.body()
    if not raw_data:
        return None
    return pickle.loads(raw_data)


class Server(ABC):
    _server: uvicorn.Server

    def __init__(self, **kwargs):
        self.agg_interval = 0.1
        self._active_clients = set()
        self._server_info = defaultdict(list)
        self._updated_server_info = False

        self._register_routes()
        self._config = uvicorn.Config(app)
        self.set_addr()
        self._disable_consol_log()

        self._quit = False
        self._last_request_time = time.time()

    def set_addr(self, host='localhost', port=18510) -> None:
        self._config.host = host
        self._config.port = port

    @staticmethod
    def enable_consol_log() -> None:
        for name in ["uvicorn", "fastapi", "uvicorn.error", "uvicorn.access"]:
            __logger = logging.getLogger(name)
            __logger.setLevel(logging.INFO)
            __logger.disabled = False

    @staticmethod
    def _disable_consol_log() -> None:
        for name in ["uvicorn", "fastapi", "uvicorn.error", "uvicorn.access"]:
            __logger = logging.getLogger(name)
            __logger.setLevel(logging.ERROR)
            __logger.disabled = True

    def update_server_info(self, name: str, value: str):
        lts = self._server_info[name][-1:]
        if [value] == lts:
            return
        else:
            self._server_info[name].append(value)
            self._updated_server_info = True

    def start(self):
        setproctitle('AlgServer')
        self._server = uvicorn.Server(self._config)
        asyncio.run(self._run_server())

    async def _run_server(self):
        await asyncio.gather(
            self._server.serve(),
            self._monitor(),
            self._aggregator()
        )

    async def _aggregator(self):
        while not self._quit:
            await asyncio.sleep(max(0.01, self.agg_interval))
            if self.num_clients == 0:
                continue
            try:
                self.aggregate()
            except Exception:
                logger.error(f"Server error: {traceback.format_exc()}")
                self.shutdown('aggregate error')

    async def _monitor(self):
        await asyncio.sleep(1)
        while not self._quit:
            self._log_server_info()
            await asyncio.sleep(.5)

    def _register_routes(self):
        @app.post("/alg-comm")
        async def alg_comm(client_pkg: ClientPackage = Depends(load_body)):
            self._last_request_time = time.time()
            try:
                server_pkg = self._handle_request(client_pkg)
            except Exception:
                logger.error(traceback.format_exc())
                self.shutdown(traceback.format_exc())
                server_pkg = None
            return Response(content=pickle.dumps(server_pkg), media_type="application/x-pickle")

    def _log_server_info(self):
        data: dict[str, list[str]] = {}
        for key, value in self._server_info.items():
            if len(value) > 0:
                data[key] = value[-1:]
        if data and self._updated_server_info:
            tab = tabulate(data, headers="keys", tablefmt="psql")
            logger.info(f"\n{'=' * 30} Saved {len(self._server_info)} clients data {'=' * 30}\n{tab}")
            self._updated_server_info = False

    @final
    def _handle_request(self, data: ClientPackage):
        if data.action == Actions.REGISTER:
            self._add_client(data.cid)
            return 'join success'
        elif data.action == Actions.QUIT:
            self._del_client(data.cid)
            return 'quit success'
        else:
            return self.handle_request(data)

    @abstractmethod
    def handle_request(self, pkg: ClientPackage) -> Any:
        ...  # pragma: no cover

    @abstractmethod
    def aggregate(self):
        ...  # pragma: no cover

    def _add_client(self, client_id):
        self._active_clients.add(client_id)
        logger.info(f"Client {client_id} join, total {self.num_clients} clients")

    def _del_client(self, client_id):
        self._active_clients.remove(client_id)
        logger.info(f"Client {client_id} quit, remain {self.num_clients} clients")
        if self.num_clients < 1:
            self.shutdown('No active clients')

    def shutdown(self, msg='no message'):
        if not self._quit:
            logger.info(f"Server shutting down ({msg})")
            self._quit = True
            self._server.should_exit = True
            self._server.force_exit = True

    @property
    def sorted_ids(self) -> list[int]:
        """
        Returns a list of client IDs sorted in ascending order.
        """
        return sorted(self._active_clients)

    @property
    def num_clients(self) -> int:
        """
        Return the number of active clients currently connected to the server.
        """
        return len(self._active_clients)
